{*************************************************************************
                PilotLogic Software House

  Package pl_GLScene
  This file is part of CodeTyphon Studio (https://www.pilotlogic.com/)

 ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 2.0
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GLScene Project
 *
 * ***** END LICENSE BLOCK *****
 ************************************************************************}

{
  Geometric objects.
}
Unit GLGeomObjects;

Interface

{$I GLScene.inc}

Uses
   Math,
   Classes,
   OpenGLTokens,
   OpenGLAdapter,
   GLScene,
   GLPersistentClasses,
   GLVectorGeometry,
   GLVectorLists,
   GLContext,
   GLObjects,
   GLSilhouette,
   GLVectorTypes,
   GLGeometryBB,
   GLRenderContextInfo,
   XOpenGL,
   GLPolynomials;

Type
  (* A Disk object that may not be complete, it can have a hole (controlled by the
    InnerRadius property) and can only be a slice (controlled by the StartAngle
    and SweepAngle properties). *)
   TGLDisk = Class(TGLQuadricObject)
   Private
      FStartAngle, FSweepAngle, FOuterRadius, FInnerRadius: Single;
      FSlices, FLoops: Integer;
      Procedure SetOuterRadius(Const aValue: Single);
      Procedure SetInnerRadius(Const aValue: Single);
      Procedure SetSlices(aValue: Integer);
      Procedure SetLoops(aValue: Integer);
      Procedure SetStartAngle(Const aValue: Single);
      Procedure SetSweepAngle(Const aValue: Single);
   Public
      Constructor Create(AOwner: TComponent); Override;
      Procedure BuildList(Var rci: TGLRenderContextInfo); Override;
      Procedure Assign(Source: TPersistent); Override;
      Function AxisAlignedDimensionsUnscaled: TVector; Override;
      Function RayCastIntersect(Const rayStart, rayVector: TVector; intersectPoint: PVector = nil; intersectNormal: PVector = nil): Boolean; Override;
   Published
      //  Allows defining a "hole" in the disk
      Property InnerRadius: Single read FInnerRadius write SetInnerRadius;
      //  Number of radial mesh subdivisions
      Property Loops: Integer read FLoops write SetLoops Default 2;
      // Outer radius for the disk. If you leave InnerRadius at 0, this is the disk radius
      Property OuterRadius: Single read FOuterRadius write SetOuterRadius;
      // Number of mesh slices. For instance, if Slices=6, your disk will look like an hexagon
      Property Slices: Integer read FSlices write SetSlices Default 16;
      Property StartAngle: Single read FStartAngle write SetStartAngle;
      Property SweepAngle: Single read FSweepAngle write SetSweepAngle;
   End;

  (*  Base class to cylinder-like objects that introduces the basic cylinder description properties.
    Be aware teh default slices and stacks make up for a high-poly cylinder,
    unless you're after high-quality lighting it is recommended to reduce the
    Stacks property to 1. *)
   TGLCylinderBase = Class(TGLQuadricObject)
   Private
      FBottomRadius: Single;
      FSlices, FStacks, FLoops: Integer;
      FHeight: Single;
   Protected
      Procedure SetBottomRadius(Const aValue: Single);
      Procedure SetHeight(Const aValue: Single);
      Procedure SetSlices(aValue: Integer);
      Procedure SetStacks(aValue: Integer);
      Procedure SetLoops(aValue: Integer);
      Function GetTopRadius: Single; Virtual;
   Public
      Constructor Create(AOwner: TComponent); Override;
      Procedure Assign(Source: TPersistent); Override;
      Function GenerateSilhouette(Const silhouetteParameters: TGLSilhouetteParameters): TGLSilhouette; Override;
   Published
      Property BottomRadius: Single read FBottomRadius write SetBottomRadius;
      Property Height: Single read FHeight write SetHeight;
      Property Slices: Integer read FSlices write SetSlices Default 16;
      Property Stacks: Integer read FStacks write SetStacks Default 4;
      {  Number of concentric rings for top/bottom disk(s). }
      Property Loops: Integer read FLoops write SetLoops Default 1;
   End;

   TConePart = (coSides, coBottom);
   TConeParts = Set Of TConePart;

   //  A cone object
   TGLCone = Class(TGLCylinderBase)
   Private
      FParts: TConeParts;
   Protected
      Procedure SetParts(aValue: TConeParts);
      Function GetTopRadius: Single; Override;
   Public
      Constructor Create(AOwner: TComponent); Override;
      Procedure Assign(Source: TPersistent); Override;
      Procedure BuildList(Var rci: TGLRenderContextInfo); Override;
      Function AxisAlignedDimensionsUnscaled: TVector; Override;
      Function RayCastIntersect(Const rayStart, rayVector: TVector; intersectPoint: PVector = nil; intersectNormal: PVector = nil): Boolean; Override;
   Published
      Property Parts: TConeParts read FParts write SetParts Default [coSides, coBottom];
   End;

   TCylinderPart = (cySides, cyBottom, cyTop);
   TCylinderParts = Set Of TCylinderPart;
   TCylinderAlignment = (caCenter, caTop, caBottom);

   {  Cylinder object, can also be used to make truncated cones }
   TGLCylinder = Class(TGLCylinderBase)
   Private
      FParts: TCylinderParts;
      FTopRadius: Single;
      FAlignment: TCylinderAlignment;
   Protected
      Procedure SetTopRadius(Const aValue: Single);
      Procedure SetParts(aValue: TCylinderParts);
      Procedure SetAlignment(val: TCylinderAlignment);
      Function GetTopRadius: Single; Override;
   Public
      Constructor Create(AOwner: TComponent); Override;
      Procedure Assign(Source: TPersistent); Override;
      Procedure BuildList(Var rci: TGLRenderContextInfo); Override;
      Function AxisAlignedDimensionsUnscaled: TVector; Override;
      Function RayCastIntersect(Const rayStart, rayVector: TVector; intersectPoint: PVector = nil; intersectNormal: PVector = nil): Boolean; Override;
      Procedure Align(Const startPoint, endPoint: TVector); Overload;
      Procedure Align(Const startObj, endObj: TGLBaseSceneObject); Overload;
      Procedure Align(Const startPoint, endPoint: TAffineVector); Overload;
   Published
      Property TopRadius: Single read FTopRadius write SetTopRadius;
      Property Parts: TCylinderParts read FParts write SetParts Default [cySides, cyBottom, cyTop];
      Property Alignment: TCylinderAlignment read FAlignment write SetAlignment Default caCenter;
   End;

   //  Capsule object, can also be used to make truncated cones
   TGLCapsule = Class(TGLSceneObject)
   Private
      FParts: TCylinderParts;
      FRadius: Single;
      FSlices: Integer;
      FStacks: Integer;
      FHeight: Single;
      FAlignment: TCylinderAlignment;
   Protected
      Procedure SetHeight(Const aValue: Single);
      Procedure SetRadius(Const aValue: Single);
      Procedure SetSlices(Const aValue: Integer);
      Procedure SetStacks(Const aValue: Integer);
      Procedure SetParts(aValue: TCylinderParts);
      Procedure SetAlignment(val: TCylinderAlignment);
   Public
      Constructor Create(AOwner: TComponent); Override;
      Procedure Assign(Source: TPersistent); Override;
      Procedure BuildList(Var rci: TGLRenderContextInfo); Override;
      Function AxisAlignedDimensionsUnscaled: TVector; Override;
      Function RayCastIntersect(Const rayStart, rayVector: TVector; intersectPoint: PVector = nil; intersectNormal: PVector = nil): Boolean; Override;
      Procedure Align(Const startPoint, endPoint: TVector); Overload;
      Procedure Align(Const startObj, endObj: TGLBaseSceneObject); Overload;
      Procedure Align(Const startPoint, endPoint: TAffineVector); Overload;
   Published
      Property Height: Single read FHeight write SetHeight;
      Property Slices: Integer read FSlices write SetSlices;
      Property Stacks: Integer read FStacks write SetStacks;
      Property Radius: Single read FRadius write SetRadius;
      Property Parts: TCylinderParts read FParts write SetParts Default [cySides, cyBottom, cyTop];
      Property Alignment: TCylinderAlignment read FAlignment write SetAlignment Default caCenter;
   End;

   TAnnulusPart = (anInnerSides, anOuterSides, anBottom, anTop);
   TAnnulusParts = Set Of TAnnulusPart;

   //  An annulus is a cylinder that can be made hollow (pipe-like)
   TGLAnnulus = Class(TGLCylinderBase)
   Private
      FParts: TAnnulusParts;
      FBottomInnerRadius: Single;
      FTopInnerRadius: Single;
      FTopRadius: Single;
   Protected
      Procedure SetTopRadius(Const aValue: Single);
      Procedure SetTopInnerRadius(Const aValue: Single);
      Procedure SetBottomInnerRadius(Const aValue: Single);
      Procedure SetParts(aValue: TAnnulusParts);
   Public
      Constructor Create(AOwner: TComponent); Override;
      Procedure Assign(Source: TPersistent); Override;
      Procedure BuildList(Var rci: TGLRenderContextInfo); Override;
      Function AxisAlignedDimensionsUnscaled: TVector; Override;
      Function RayCastIntersect(Const rayStart, rayVector: TVector; intersectPoint: PVector = nil; intersectNormal: PVector = nil): Boolean; Override;
   Published
      Property BottomInnerRadius: Single read FBottomInnerRadius write SetBottomInnerRadius;
      Property TopInnerRadius: Single read FTopInnerRadius write SetTopInnerRadius;
      Property TopRadius: Single read FTopRadius write SetTopRadius;
      Property Parts: TAnnulusParts read FParts write SetParts Default [anInnerSides, anOuterSides, anBottom, anTop];
   End;

   TTorusPart = (toSides, toStartDisk, toStopDisk);
   TTorusParts = Set Of TTorusPart;

   //  A Torus object
   TGLTorus = Class(TGLSceneObject)
   Private
      FParts: TTorusParts;
      FRings, FSides: Cardinal;
      FStartAngle, FStopAngle: Single;
      FMinorRadius, FMajorRadius: Single;
      FMesh: Array Of Array Of TVertexRec;
   Protected
      Procedure SetMajorRadius(Const aValue: Single);
      Procedure SetMinorRadius(Const aValue: Single);
      Procedure SetRings(aValue: Cardinal);
      Procedure SetSides(aValue: Cardinal);
      Procedure SetStartAngle(Const aValue: Single);
      Procedure SetStopAngle(Const aValue: Single);
      Procedure SetParts(aValue: TTorusParts);
   Public
      Constructor Create(AOwner: TComponent); Override;
      Procedure BuildList(Var rci: TGLRenderContextInfo); Override;
      Function AxisAlignedDimensionsUnscaled: TVector; Override;
      Function RayCastIntersect(Const rayStart, rayVector: TVector; intersectPoint: PVector = nil; intersectNormal: PVector = nil): Boolean; Override;
   Published
      Property MajorRadius: Single read FMajorRadius write SetMajorRadius;
      Property MinorRadius: Single read FMinorRadius write SetMinorRadius;
      Property Rings: Cardinal read FRings write SetRings Default 25;
      Property Sides: Cardinal read FSides write SetSides Default 15;
      Property StartAngle: Single read FStartAngle write SetStartAngle;
      Property StopAngle: Single read FStopAngle write SetStopAngle;
      Property Parts: TTorusParts read FParts write SetParts Default [toSides];
   End;

   TArrowLinePart = (alLine, alTopArrow, alBottomArrow);
   TArrowLineParts = Set Of TArrowLinePart;

   TArrowHeadStackingStyle = (ahssStacked, ahssCentered, ahssIncluded);

  { Draws an arrowhead (cylinder + cone).
    The arrow head is a cone that shares the attributes of the cylinder
    (ie stacks/slices, materials etc). Seems to work ok.
    This is useful for displaying a vector based field (eg velocity) or
    other arrows that might be required.
    By default the bottom arrow is off }
   TGLArrowLine = Class(TGLCylinderBase)
   Private
      FParts: TArrowLineParts;
      FTopRadius: Single;
      fTopArrowHeadHeight: Single;
      fTopArrowHeadRadius: Single;
      fBottomArrowHeadHeight: Single;
      fBottomArrowHeadRadius: Single;
      FHeadStackingStyle: TArrowHeadStackingStyle;
   Protected
      Procedure SetTopRadius(Const aValue: Single);
      Procedure SetTopArrowHeadHeight(Const aValue: Single);
      Procedure SetTopArrowHeadRadius(Const aValue: Single);
      Procedure SetBottomArrowHeadHeight(Const aValue: Single);
      Procedure SetBottomArrowHeadRadius(Const aValue: Single);
      Procedure SetParts(aValue: TArrowLineParts);
      Procedure SetHeadStackingStyle(Const val: TArrowHeadStackingStyle);
   Public
      Constructor Create(AOwner: TComponent); Override;
      Procedure BuildList(Var rci: TGLRenderContextInfo); Override;
      Procedure Assign(Source: TPersistent); Override;
   Published
      Property TopRadius: Single read FTopRadius write SetTopRadius;
      Property HeadStackingStyle: TArrowHeadStackingStyle read FHeadStackingStyle write SetHeadStackingStyle Default ahssStacked;
      Property Parts: TArrowLineParts read FParts write SetParts Default [alLine, alTopArrow];
      Property TopArrowHeadHeight: Single read fTopArrowHeadHeight write SetTopArrowHeadHeight;
      Property TopArrowHeadRadius: Single read fTopArrowHeadRadius write SetTopArrowHeadRadius;
      Property BottomArrowHeadHeight: Single read fBottomArrowHeadHeight write SetBottomArrowHeadHeight;
      Property BottomArrowHeadRadius: Single read fBottomArrowHeadRadius write SetBottomArrowHeadRadius;
   End;

   TArrowArcPart = (aaArc, aaTopArrow, aaBottomArrow);
   TArrowArcParts = Set Of TArrowArcPart;

  (* Draws an arrowhead (Sliced Torus + cone).
    The arrow head is a cone that shares the attributes of the Torus
    (ie stacks/slices, materials etc).
    This is useful for displaying a movement (eg twist) or
    other arc arrows that might be required.
    By default the bottom arrow is off *)
   TGLArrowArc = Class(TGLCylinderBase)
   Private
      fArcRadius: Single;
      FStartAngle: Single;
      FStopAngle: Single;
      FParts: TArrowArcParts;
      FTopRadius: Single;
      fTopArrowHeadHeight: Single;
      fTopArrowHeadRadius: Single;
      fBottomArrowHeadHeight: Single;
      fBottomArrowHeadRadius: Single;
      FHeadStackingStyle: TArrowHeadStackingStyle;
      FMesh: Array Of Array Of TVertexRec;
   Protected
      Procedure SetArcRadius(Const aValue: Single);
      Procedure SetStartAngle(Const aValue: Single);
      Procedure SetStopAngle(Const aValue: Single);
      Procedure SetTopRadius(Const aValue: Single);
      Procedure SetTopArrowHeadHeight(Const aValue: Single);
      Procedure SetTopArrowHeadRadius(Const aValue: Single);
      Procedure SetBottomArrowHeadHeight(Const aValue: Single);
      Procedure SetBottomArrowHeadRadius(Const aValue: Single);
      Procedure SetParts(aValue: TArrowArcParts);
      Procedure SetHeadStackingStyle(Const val: TArrowHeadStackingStyle);
   Public
      Constructor Create(AOwner: TComponent); Override;
      Procedure BuildList(Var rci: TGLRenderContextInfo); Override;
      Procedure Assign(Source: TPersistent); Override;
   Published
      Property ArcRadius: Single read fArcRadius write SetArcRadius;
      Property StartAngle: Single read FStartAngle write SetStartAngle;
      Property StopAngle: Single read FStopAngle write SetStopAngle;
      Property TopRadius: Single read FTopRadius write SetTopRadius;
      Property HeadStackingStyle: TArrowHeadStackingStyle read FHeadStackingStyle write SetHeadStackingStyle Default ahssStacked;
      Property Parts: TArrowArcParts read FParts write SetParts Default [aaArc, aaTopArrow];
      Property TopArrowHeadHeight: Single read fTopArrowHeadHeight write SetTopArrowHeadHeight;
      Property TopArrowHeadRadius: Single read fTopArrowHeadRadius write SetTopArrowHeadRadius;
      Property BottomArrowHeadHeight: Single read fBottomArrowHeadHeight write SetBottomArrowHeadHeight;
      Property BottomArrowHeadRadius: Single read fBottomArrowHeadRadius write SetBottomArrowHeadRadius;
   End;

   TPolygonPart = (ppTop, ppBottom);
   TPolygonParts = Set Of TPolygonPart;

  (* A basic polygon object.
    The curve is described by the Nodes and SplineMode properties, should be
    planar and is automatically tessellated.
    Texture coordinates are deduced from X and Y coordinates only.
    This object allows only for polygons described by a single curve, if you
    need "complex polygons" with holes, patches and cutouts, see GLMultiPolygon. *)
   TGLPolygon = Class(TGLPolygonBase)
   Private
      FParts: TPolygonParts;
   Protected
      Procedure SetParts(Const val: TPolygonParts);
   Public
      Constructor Create(AOwner: TComponent); Override;
      Destructor Destroy; Override;
      Procedure Assign(Source: TPersistent); Override;
      Procedure BuildList(Var rci: TGLRenderContextInfo); Override;
   Published
    {  Parts of polygon.
      The 'top' of the polygon is the position were the curve describing
      the polygon spin counter-clockwise (i.e. right handed convention). }
      Property Parts: TPolygonParts read FParts write SetParts Default [ppTop, ppBottom];
   End;

   TFrustrumPart = (fpTop, fpBottom, fpFront, fpBack, fpLeft, fpRight);
   TFrustrumParts = Set Of TFrustrumPart;

Const
   cAllFrustrumParts = [fpTop, fpBottom, fpFront, fpBack, fpLeft, fpRight];

Type
  { A frustrum is a pyramid with the top chopped off.
    The height of the imaginary pyramid is ApexHeight, the height of the
    frustrum is Height. If ApexHeight and Height are the same, the frustrum
    degenerates into a pyramid.
    Height cannot be greater than ApexHeight. }
   TGLFrustrum = Class(TGLSceneObject)
   Private
      FApexHeight, FBaseDepth, FBaseWidth, FHeight: Single;
      FParts: TFrustrumParts;
      FNormalDirection: TNormalDirection;
      Procedure SetApexHeight(Const aValue: Single);
      Procedure SetBaseDepth(Const aValue: Single);
      Procedure SetBaseWidth(Const aValue: Single);
      Procedure SetHeight(Const aValue: Single);
      Procedure SetParts(aValue: TFrustrumParts);
      Procedure SetNormalDirection(aValue: TNormalDirection);
   Protected
      Procedure DefineProperties(Filer: TFiler); Override;
      Procedure ReadData(Stream: TStream);
      Procedure WriteData(Stream: TStream);
   Public
      Constructor Create(AOwner: TComponent); Override;
      Procedure BuildList(Var rci: TGLRenderContextInfo); Override;
      Procedure Assign(Source: TPersistent); Override;
      Function TopDepth: Single;
      Function TopWidth: Single;
      Function AxisAlignedBoundingBoxUnscaled: TAABB;
      Function AxisAlignedDimensionsUnscaled: TVector; Override;
   Published
      Property ApexHeight: Single read FApexHeight write SetApexHeight Stored False;
      Property BaseDepth: Single read FBaseDepth write SetBaseDepth Stored False;
      Property BaseWidth: Single read FBaseWidth write SetBaseWidth Stored False;
      Property Height: Single read FHeight write SetHeight Stored False;
      Property NormalDirection: TNormalDirection read FNormalDirection write SetNormalDirection Default ndOutside;
      Property Parts: TFrustrumParts read FParts write SetParts Default cAllFrustrumParts;
   End;


Implementation



// ------------------
// ------------------ TGLDisk ------------------
// ------------------

Constructor TGLDisk.Create(AOwner: TComponent);
Begin
   Inherited Create(AOwner);
   FOuterRadius := 0.5;
   FInnerRadius := 0;
   FSlices := 16;
   FLoops := 2;
   FStartAngle := 0;
   FSweepAngle := 360;
End;

Procedure TGLDisk.BuildList(Var rci: TGLRenderContextInfo);
Var
   quadric: PGLUquadric;
   //  quadric: GLUquadricObj;  from Winapi.OpenGL
Begin
   quadric := gluNewQuadric();
   SetupQuadricParams(quadric);
   gluPartialDisk(quadric, FInnerRadius, FOuterRadius, FSlices, FLoops,
      FStartAngle, FSweepAngle);
   gluDeleteQuadric(quadric);
End;

Procedure TGLDisk.SetOuterRadius(Const aValue: Single);
Begin
   If aValue <> FOuterRadius Then
   Begin
      FOuterRadius := aValue;
      StructureChanged;
   End;
End;

Procedure TGLDisk.SetInnerRadius(Const aValue: Single);
Begin
   If aValue <> FInnerRadius Then
   Begin
      FInnerRadius := aValue;
      StructureChanged;
   End;
End;

Procedure TGLDisk.SetSlices(aValue: Integer);
Begin
   If aValue <> FSlices Then
   Begin
      FSlices := aValue;
      StructureChanged;
   End;
End;

Procedure TGLDisk.SetLoops(aValue: Integer);
Begin
   If aValue <> FLoops Then
   Begin
      FLoops := aValue;
      StructureChanged;
   End;
End;

Procedure TGLDisk.SetStartAngle(Const aValue: Single);
Begin
   If aValue <> FStartAngle Then
   Begin
      FStartAngle := aValue;
      StructureChanged;
   End;
End;

Procedure TGLDisk.SetSweepAngle(Const aValue: Single);
Begin
   If aValue <> FSweepAngle Then
   Begin
      FSweepAngle := aValue;
      StructureChanged;
   End;
End;

Procedure TGLDisk.Assign(Source: TPersistent);
Begin
   If Assigned(Source) And (Source Is TGLDisk) Then
   Begin
      FOuterRadius := TGLDisk(Source).FOuterRadius;
      FInnerRadius := TGLDisk(Source).FInnerRadius;
      FSlices := TGLDisk(Source).FSlices;
      FLoops := TGLDisk(Source).FLoops;
      FStartAngle := TGLDisk(Source).FStartAngle;
      FSweepAngle := TGLDisk(Source).FSweepAngle;
   End;
   Inherited Assign(Source);
End;

Function TGLDisk.AxisAlignedDimensionsUnscaled: TVector;
Var
   r: Single;
Begin
   r := Abs(FOuterRadius);
   Result := VectorMake(r, r, 0);
End;

Function TGLDisk.RayCastIntersect(Const rayStart, rayVector: TVector; intersectPoint: PVector = nil; intersectNormal: PVector = nil): Boolean;
Var
   ip: TVector;
   d: Single;
   angle, beginAngle, endAngle: Single;
   localIntPoint: TVector;
Begin
   Result := False;
   If SweepAngle > 0 Then
      If RayCastPlaneIntersect(rayStart, rayVector, AbsolutePosition, AbsoluteDirection, @ip) Then
      Begin
         If Assigned(intersectPoint) Then
            SetVector(intersectPoint^, ip);
         localIntPoint := AbsoluteToLocal(ip);
         d := VectorNorm(localIntPoint);
         If (d >= Sqr(InnerRadius)) And (d <= Sqr(OuterRadius)) Then
         Begin
            If SweepAngle >= 360 Then
               Result := True
            Else Begin
               // arctan2 returns results between -pi and +pi, we want between 0 and 360
               angle := 180 / pi * ArcTan2(localIntPoint.X, localIntPoint.Y);
               If angle < 0 Then
                  angle := angle + 360;
               // we also want StartAngle and StartAngle+SweepAngle to be in this range
               beginAngle := Trunc(StartAngle) Mod 360;
               endAngle := Trunc(StartAngle + SweepAngle) Mod 360;
               // If beginAngle>endAngle then area crosses the boundary from 360=>0 degrees
               // therefore have 2 valid regions  (beginAngle to 360) & (0 to endAngle)
               // otherwise just 1 valid region (beginAngle to endAngle)
               If beginAngle > endAngle Then
               Begin
                  If (angle > beginAngle) Or (angle < endAngle) Then
                     Result := True;
               End Else
               If (angle > beginAngle) And (angle < endAngle) Then
                  Result := True;
            End;
         End;
      End;
   If Result = True Then
      If Assigned(intersectNormal) Then
         SetVector(intersectNormal^, AbsoluteUp);
End;

// ------------------
// ------------------ TGLCylinderBase ------------------
// ------------------

Constructor TGLCylinderBase.Create(AOwner: TComponent);
Begin
   Inherited Create(AOwner);
   FBottomRadius := 0.5;
   FHeight := 1;
   FSlices := 16;
   FStacks := 4;
   FLoops := 1;
End;

Procedure TGLCylinderBase.SetBottomRadius(Const aValue: Single);
Begin
   If aValue <> FBottomRadius Then
   Begin
      FBottomRadius := aValue;
      StructureChanged;
   End;
End;

Function TGLCylinderBase.GetTopRadius: Single;
Begin
   Result := FBottomRadius;
End;

Procedure TGLCylinderBase.SetHeight(Const aValue: Single);
Begin
   If aValue <> FHeight Then
   Begin
      FHeight := aValue;
      StructureChanged;
   End;
End;

Procedure TGLCylinderBase.SetSlices(aValue: Integer);
Begin
   If aValue <> FSlices Then
   Begin
      FSlices := aValue;
      StructureChanged;
   End;
End;

Procedure TGLCylinderBase.SetStacks(aValue: Integer);
Begin
   If aValue <> FStacks Then
   Begin
      FStacks := aValue;
      StructureChanged;
   End;
End;

Procedure TGLCylinderBase.SetLoops(aValue: Integer);
Begin
   If (aValue >= 1) And (aValue <> FLoops) Then
   Begin
      FLoops := aValue;
      StructureChanged;
   End;
End;

Procedure TGLCylinderBase.Assign(Source: TPersistent);
Begin
   If Assigned(Source) And (Source Is TGLCylinderBase) Then
   Begin
      FBottomRadius := TGLCylinderBase(Source).FBottomRadius;
      FSlices := TGLCylinderBase(Source).FSlices;
      FStacks := TGLCylinderBase(Source).FStacks;
      FLoops := TGLCylinderBase(Source).FLoops;
      FHeight := TGLCylinderBase(Source).FHeight;
   End;
   Inherited Assign(Source);
End;


Function TGLCylinderBase.GenerateSilhouette(Const silhouetteParameters: TGLSilhouetteParameters): TGLSilhouette;
Var
   connectivity: TGLConnectivity;
   sil: TGLSilhouette;
   ShadowSlices: Integer;
   i: Integer;
   p: Array [0 .. 3] Of TVector3f;
   PiDivSlices: Single;
   a1, a2: Single;
   c1, c2: TVector3f;
   cosa1, cosa2, sina1, sina2: Single;
   HalfHeight: Single;
   ShadowTopRadius: Single;
Begin
   connectivity := TGLConnectivity.Create(True);

   ShadowSlices := FSlices Div 1;

   If FSlices < 5 Then
      FSlices := 5;

   PiDivSlices := 2 * pi / ShadowSlices;

   a1 := 0;

   // Is this a speed improvement or just a waste of code?
   HalfHeight := FHeight / 2;

   MakeVector(c1, 0, -HalfHeight, 0);
   MakeVector(c2, 0, HalfHeight, 0);

   ShadowTopRadius := GetTopRadius;

   For i := 0 To ShadowSlices - 1 Do
   Begin
      a2 := a1 + PiDivSlices;

      // Is this a speed improvement or just a waste of code?
      cosa1 := cos(a1);
      cosa2 := cos(a2);
      sina1 := sin(a1);
      sina2 := sin(a2);

      // Generate the four "corners";
      // Bottom corners
      MakeVector(p[0], FBottomRadius * sina2, -HalfHeight, FBottomRadius * cosa2);
      MakeVector(p[1], FBottomRadius * sina1, -HalfHeight, FBottomRadius * cosa1);

      // Top corners
      MakeVector(p[2], ShadowTopRadius * sina1, HalfHeight,
         ShadowTopRadius * cosa1);
      MakeVector(p[3], ShadowTopRadius * sina2, HalfHeight,
         ShadowTopRadius * cosa2); // }

      // This should be optimized to use AddIndexedFace, because this method
      // searches for each of the vertices and adds them or re-uses them.

      // Skin
      connectivity.AddFace(p[2], p[1], p[0]);
      connectivity.AddFace(p[3], p[2], p[0]);

      // Sides / caps
      connectivity.AddFace(c1, p[0], p[1]);
      connectivity.AddFace(p[2], p[3], c2);

      a1 := a1 + PiDivSlices;
   End;

   sil := nil;
   connectivity.CreateSilhouette(silhouetteParameters, sil, False);

   Result := sil;

   connectivity.Free;
End;

// ------------------
// ------------------ TGLCone ------------------
// ------------------

Constructor TGLCone.Create(AOwner: TComponent);
Begin
   Inherited Create(AOwner);
   FParts := [coSides, coBottom];
End;

Procedure TGLCone.BuildList(Var rci: TGLRenderContextInfo);
Var
   quadric: PGLUquadricObj;
Begin
   gl.PushMatrix;
   quadric := gluNewQuadric();
   SetupQuadricParams(quadric);
   gl.Rotated(-90, 1, 0, 0);
   gl.Translatef(0, 0, -FHeight * 0.5);
   If coSides In FParts Then
      gluCylinder(quadric, BottomRadius, 0, Height, Slices, Stacks);
   If coBottom In FParts Then
   Begin
      // top of a disk is defined as outside
      SetInvertedQuadricOrientation(quadric);
      gluDisk(quadric, 0, BottomRadius, Slices, FLoops);
   End;
   gluDeleteQuadric(quadric);
   gl.PopMatrix;
End;

Procedure TGLCone.SetParts(aValue: TConeParts);
Begin
   If aValue <> FParts Then
   Begin
      FParts := aValue;
      StructureChanged;
   End;
End;

Procedure TGLCone.Assign(Source: TPersistent);
Begin
   If Assigned(Source) And (Source Is TGLCone) Then
   Begin
      FParts := TGLCone(Source).FParts;
   End;
   Inherited Assign(Source);
End;

Function TGLCone.AxisAlignedDimensionsUnscaled: TVector;
Var
   r: Single;
Begin
   r := Abs(FBottomRadius);
   Result := VectorMake(r { *Scale.DirectX }, 0.5 * FHeight { *Scale.DirectY }, r { *Scale.DirectZ });
End;

Function TGLCone.GetTopRadius: Single;
Begin
   Result := 0;
End;

Function TGLCone.RayCastIntersect(Const rayStart, rayVector: TVector; intersectPoint: PVector = nil; intersectNormal: PVector = nil): Boolean;
Var
   ip, localRayStart, localRayVector: TVector;
   poly: Array [0 .. 2] Of Double;
   roots: TDoubleArray;
   minRoot: Double;
   d, t, hconst: Single;
Begin
   Result := False;
   localRayStart := AbsoluteToLocal(rayStart);
   localRayVector := VectorNormalize(AbsoluteToLocal(rayVector));

   If coBottom In Parts Then
   Begin
      // bottom can only be raycast from beneath
      If localRayStart.Y < -FHeight * 0.5 Then
      Begin
         If RayCastPlaneIntersect(localRayStart, localRayVector, PointMake(0, -FHeight * 0.5, 0), YHmgVector, @ip) Then
         Begin
            d := VectorNorm(ip.X, ip.Z);
            If (d <= Sqr(BottomRadius)) Then
            Begin
               Result := True;
               If Assigned(intersectPoint) Then
                  SetVector(intersectPoint^, LocalToAbsolute(ip));
               If Assigned(intersectNormal) Then
                  SetVector(intersectNormal^, VectorNegate(AbsoluteUp));
               Exit;
            End;
         End;
      End;
   End;
   If coSides In Parts Then
   Begin
      hconst := -Sqr(BottomRadius) / Sqr(Height);
      // intersect against infinite cones (in positive and negative direction)
      poly[0] := Sqr(localRayStart.X) + hconst * Sqr(localRayStart.Y - 0.5 * FHeight) + Sqr(localRayStart.Z);
      poly[1] := 2 * (localRayStart.X * localRayVector.X + hconst * (localRayStart.Y - 0.5 * FHeight) * localRayVector.Y + localRayStart.Z * localRayVector.Z);
      poly[2] := Sqr(localRayVector.X) + hconst * Sqr(localRayVector.Y) + Sqr(localRayVector.Z);
      SetLength(roots, 0);
      roots := SolveQuadric(@poly);
      If MinPositiveCoef(roots, minRoot) Then
      Begin
         t := minRoot;
         ip := VectorCombine(localRayStart, localRayVector, 1, t);
         // check that intersection with infinite cone is within the range we want
         If (ip.Y > -FHeight * 0.5) And (ip.Y < FHeight * 0.5) Then
         Begin
            Result := True;
            If Assigned(intersectPoint) Then
               intersectPoint^ := LocalToAbsolute(ip);
            If Assigned(intersectNormal) Then
            Begin
               ip.Y := hconst * (ip.Y - 0.5 * Height);
               ip.W := 0;
               NormalizeVector(ip);
               intersectNormal^ := LocalToAbsolute(ip);
            End;
         End;
      End;
   End;
End;

// ------------------
// ------------------ TGLCylinder ------------------
// ------------------

Constructor TGLCylinder.Create(AOwner: TComponent);
Begin
   Inherited Create(AOwner);
   FTopRadius := 0.5;
   FParts := [cySides, cyBottom, cyTop];
   FAlignment := caCenter;
End;

Procedure TGLCylinder.BuildList(Var rci: TGLRenderContextInfo);
Var
   quadric: PGLUquadricObj;
Begin
   gl.PushMatrix;
   quadric := gluNewQuadric;
   SetupQuadricParams(quadric);
   gl.Rotatef(-90, 1, 0, 0);
   Case Alignment Of
      caTop:
         gl.Translatef(0, 0, -FHeight);
      caBottom:
         ;
      Else // caCenter
         gl.Translatef(0, 0, -FHeight * 0.5);
   End;
   If cySides In FParts Then
      gluCylinder(quadric, FBottomRadius, FTopRadius, FHeight, FSlices, FStacks);
   If cyTop In FParts Then
   Begin
      gl.PushMatrix;
      gl.Translatef(0, 0, FHeight);
      gluDisk(Quadric, 0, FTopRadius, FSlices, FLoops);
      gl.PopMatrix;
   End;
   If cyBottom In FParts Then
   Begin
      // swap quadric orientation because top of a disk is defined as outside
      SetInvertedQuadricOrientation(quadric);
      gluDisk(quadric, 0, FBottomRadius, FSlices, FLoops);
   End;
   gluDeleteQuadric(Quadric);
   gl.PopMatrix;
End;

Procedure TGLCylinder.SetTopRadius(Const aValue: Single);
Begin
   If aValue <> FTopRadius Then
   Begin
      FTopRadius := aValue;
      StructureChanged;
   End;
End;

Function TGLCylinder.GetTopRadius: Single;
Begin
   Result := FTopRadius;
End;

Procedure TGLCylinder.SetParts(aValue: TCylinderParts);
Begin
   If aValue <> FParts Then
   Begin
      FParts := aValue;
      StructureChanged;
   End;
End;

Procedure TGLCylinder.SetAlignment(val: TCylinderAlignment);
Begin
   If val <> FAlignment Then
   Begin
      FAlignment := val;
      StructureChanged;
   End;
End;

Procedure TGLCylinder.Assign(Source: TPersistent);
Begin
   If Assigned(Source) And (Source Is TGLCylinder) Then
   Begin
      FParts := TGLCylinder(Source).FParts;
      FTopRadius := TGLCylinder(Source).FTopRadius;
   End;
   Inherited Assign(Source);
End;

Function TGLCylinder.AxisAlignedDimensionsUnscaled: TVector;
Var
   r, r1: Single;
Begin
   r := Abs(FBottomRadius);
   r1 := Abs(FTopRadius);
   If r1 > r Then
      r := r1;
   Result := VectorMake(r, 0.5 * FHeight, r);
   // ScaleVector(Result, Scale.AsVector);
End;

Function TGLCylinder.RayCastIntersect(Const rayStart, rayVector: TVector; intersectPoint: PVector = nil; intersectNormal: PVector = nil): Boolean;
Const
   cOne: Single = 1;
Var
   locRayStart, locRayVector, ip: TVector;
   poly: Array [0 .. 2] Of Double;
   roots: TDoubleArray;
   minRoot: Double;
   t, tr2, invRayVector1, hTop, hBottom: Single;
   tPlaneMin, tPlaneMax: Single;
Begin
   Result := False;
   locRayStart := AbsoluteToLocal(rayStart);
   locRayVector := AbsoluteToLocal(rayVector);

   Case Alignment Of
      caTop:
      Begin
         hTop := 0;
         hBottom := -Height;
      End;
      caBottom:
      Begin
         hTop := Height;
         hBottom := 0;
      End;
      Else // caCenter
         hTop := Height * 0.5;
         hBottom := -hTop;
   End;

   If locRayVector.Y = 0 Then
   Begin
      // intersect if ray shot through the top/bottom planes
      If (locRayStart.X > hTop) Or (locRayStart.X < hBottom) Then
         Exit;
      tPlaneMin := -1E99;
      tPlaneMax := 1E99;
   End Else Begin
      invRayVector1 := cOne / locRayVector.Y;
      tr2 := Sqr(TopRadius);

      // compute intersection with topPlane
      t := (hTop - locRayStart.Y) * invRayVector1;
      If (t > 0) And (cyTop In Parts) Then
      Begin
         ip.X := locRayStart.X + t * locRayVector.X;
         ip.Z := locRayStart.Z + t * locRayVector.Z;
         If Sqr(ip.X) + Sqr(ip.Z) <= tr2 Then
         Begin
            // intersect with top plane
            If Assigned(intersectPoint) Then
               intersectPoint^ := LocalToAbsolute(VectorMake(ip.X, hTop, ip.Z, 1));
            If Assigned(intersectNormal) Then
               intersectNormal^ := LocalToAbsolute(YHmgVector);
            Result := True;
         End;
      End;
      tPlaneMin := t;
      tPlaneMax := t;
      // compute intersection with bottomPlane
      t := (hBottom - locRayStart.Y) * invRayVector1;
      If (t > 0) And (cyBottom In Parts) Then
      Begin
         ip.X := locRayStart.X + t * locRayVector.X;
         ip.Z := locRayStart.Z + t * locRayVector.Z;
         If (t < tPlaneMin) Or (Not (cyTop In Parts)) Then
         Begin
            If Sqr(ip.X) + Sqr(ip.Z) <= tr2 Then
            Begin
               // intersect with top plane
               If Assigned(intersectPoint) Then
                  intersectPoint^ := LocalToAbsolute(VectorMake(ip.X, hBottom, ip.Z, 1));
               If Assigned(intersectNormal) Then
                  intersectNormal^ := LocalToAbsolute(VectorNegate(YHmgVector));
               Result := True;
            End;
         End;
      End;
      If t < tPlaneMin Then
         tPlaneMin := t;
      If t > tPlaneMax Then
         tPlaneMax := t;
   End;
   If cySides In Parts Then
   Begin
      // intersect against cylinder infinite cylinder
      poly[0] := Sqr(locRayStart.X) + Sqr(locRayStart.Z) - Sqr(TopRadius);
      poly[1] := 2 * (locRayStart.X * locRayVector.X + locRayStart.Z * locRayVector.Z);
      poly[2] := Sqr(locRayVector.X) + Sqr(locRayVector.Z);
      roots := SolveQuadric(@poly);
      If MinPositiveCoef(roots, minRoot) Then
      Begin
         t := minRoot;
         If (t >= tPlaneMin) And (t < tPlaneMax) Then
         Begin
            If Assigned(intersectPoint) Or Assigned(intersectNormal) Then
            Begin
               ip := VectorCombine(locRayStart, locRayVector, 1, t);
               If Assigned(intersectPoint) Then
                  intersectPoint^ := LocalToAbsolute(ip);
               If Assigned(intersectNormal) Then
               Begin
                  ip.Y := 0;
                  ip.W := 0;
                  intersectNormal^ := LocalToAbsolute(ip);
               End;
            End;
            Result := True;
         End;
      End;
   End Else
      SetLength(roots, 0);
End;

Procedure TGLCylinder.Align(Const startPoint, endPoint: TVector);
Var
   dir: TAffineVector;
Begin
   AbsolutePosition := startPoint;
   VectorSubtract(endPoint, startPoint, dir);
   If Parent <> nil Then
      dir := Parent.AbsoluteToLocal(dir);
   Up.AsAffineVector := dir;
   Height := VectorLength(dir);
   Lift(Height * 0.5);
   Alignment := caCenter;
End;

Procedure TGLCylinder.Align(Const startObj, endObj: TGLBaseSceneObject);
Begin
   Align(startObj.AbsolutePosition, endObj.AbsolutePosition);
End;

Procedure TGLCylinder.Align(Const startPoint, endPoint: TAffineVector);
Begin
   Align(PointMake(startPoint), PointMake(endPoint));
End;

// ------------------
// ------------------ TGLCapsule ------------------
// ------------------

Constructor TGLCapsule.Create(AOwner: TComponent);
Begin
   Inherited Create(AOwner);
   FHeight := 1;
   FRadius := 0.5;
   FSlices := 4;
   FStacks := 4;
   FParts := [cySides, cyBottom, cyTop];
   FAlignment := caCenter;
End;


Procedure TGLCapsule.BuildList(Var rci: TGLRenderContextInfo);
Var
   i, j, n: Integer;
   start_nx2: Single;
   start_ny2: Single;
   tmp, nx, ny, nz, start_nx, start_ny, a, ca, sa, l: Single;
   nx2, ny2, nz2: Single;
Begin
   gl.PushMatrix;
   gl.Rotatef(-90, 0, 0, 1);
   Case Alignment Of
      caTop:
         gl.Translatef(0, 0, FHeight + 1);
      caBottom:
         gl.Translatef(0, 0, -FHeight);
      Else // caCenter
         gl.Translatef(0, 0, 0.5);
   End;
   n := FSlices * FStacks;
   l := FHeight;
   l := l * 0.5;
   a := (pi * 2.0) / n;
   sa := sin(a);
   ca := cos(a);
   ny := 0;
   nz := 1;
   If cySides In FParts Then
   Begin
      gl.Begin_(GL_TRIANGLE_STRIP);
      For i := 0 To n Do
      Begin
         gl.Normal3d(ny, nz, 0);
         gl.TexCoord2f(i / n, 1);
         gl.Vertex3d(ny * FRadius, nz * FRadius, l - 0.5);
         gl.Normal3d(ny, nz, 0);
         gl.TexCoord2f(i / n, 0);
         gl.Vertex3d(ny * FRadius, nz * FRadius, -l - 0.5);
         tmp := ca * ny - sa * nz;
         nz := sa * ny + ca * nz;
         ny := tmp;
      End;
      gl.End_();
   End;

   If cyTop In FParts Then
   Begin
      start_nx := 0;
      start_ny := 1;
      For j := 0 To (n Div FStacks) Do
      Begin
         start_nx2 := ca * start_nx + sa * start_ny;
         start_ny2 := -sa * start_nx + ca * start_ny;
         nx := start_nx;
         ny := start_ny;
         nz := 0;
         nx2 := start_nx2;
         ny2 := start_ny2;
         nz2 := 0;
         gl.PushMatrix;
         gl.Translatef(0, 0, -0.5);
         gl.Begin_(GL_TRIANGLE_STRIP);
         For i := 0 To n Do
         Begin
            gl.Normal3d(ny2, nz2, nx2);
            gl.TexCoord2f(i / n, j / n);
            gl.Vertex3d(ny2 * FRadius, nz2 * FRadius, l + nx2 * FRadius);
            gl.Normal3d(ny, nz, nx);
            gl.TexCoord2f(i / n, (j - 1) / n);
            gl.Vertex3d(ny * FRadius, nz * FRadius, l + nx * FRadius);
            tmp := ca * ny - sa * nz;
            nz := sa * ny + ca * nz;
            ny := tmp;
            tmp := ca * ny2 - sa * nz2;
            nz2 := sa * ny2 + ca * nz2;
            ny2 := tmp;
         End;
         gl.End_();
         gl.PopMatrix;
         start_nx := start_nx2;
         start_ny := start_ny2;
      End;
   End;

   If cyBottom In FParts Then
   Begin
      start_nx := 0;
      start_ny := 1;
      For j := 0 To (n Div FStacks) Do
      Begin
         start_nx2 := ca * start_nx - sa * start_ny;
         start_ny2 := sa * start_nx + ca * start_ny;
         nx := start_nx;
         ny := start_ny;
         nz := 0;
         nx2 := start_nx2;
         ny2 := start_ny2;
         nz2 := 0;
         gl.PushMatrix;
         gl.Translatef(0, 0, -0.5);
         gl.Begin_(GL_TRIANGLE_STRIP);
         For i := 0 To n Do
         Begin
            gl.Normal3d(ny, nz, nx);
            gl.TexCoord2f(i / n, (j - 1) / n);
            gl.Vertex3d(ny * FRadius, nz * FRadius, -l + nx * FRadius);
            gl.Normal3d(ny2, nz2, nx2);
            gl.TexCoord2f(i / n, j / n);
            gl.Vertex3d(ny2 * FRadius, nz2 * FRadius, -l + nx2 * FRadius);
            tmp := ca * ny - sa * nz;
            nz := sa * ny + ca * nz;
            ny := tmp;
            tmp := ca * ny2 - sa * nz2;
            nz2 := sa * ny2 + ca * nz2;
            ny2 := tmp;
         End;
         gl.End_();
         gl.PopMatrix;
         start_nx := start_nx2;
         start_ny := start_ny2;
      End;
   End;
   gl.PopMatrix;
End;

Procedure TGLCapsule.SetHeight(Const aValue: Single);
Begin
   If aValue <> FHeight Then
   Begin
      FHeight := aValue;
      StructureChanged;
   End;
End;

Procedure TGLCapsule.SetRadius(Const aValue: Single);
Begin
   If aValue <> FRadius Then
   Begin
      FRadius := aValue;
      StructureChanged;
   End;
End;

Procedure TGLCapsule.SetSlices(Const aValue: Integer);
Begin
   If aValue <> FSlices Then
   Begin
      FSlices := aValue;
      StructureChanged;
   End;
End;

Procedure TGLCapsule.SetStacks(Const aValue: Integer);
Begin
   If aValue <> FStacks Then
   Begin
      FStacks := aValue;
      StructureChanged;
   End;
End;

Procedure TGLCapsule.SetParts(aValue: TCylinderParts);
Begin
   If aValue <> FParts Then
   Begin
      FParts := aValue;
      StructureChanged;
   End;
End;

Procedure TGLCapsule.SetAlignment(val: TCylinderAlignment);
Begin
   If val <> FAlignment Then
   Begin
      FAlignment := val;
      StructureChanged;
   End;
End;

Procedure TGLCapsule.Assign(Source: TPersistent);
Begin
   If Assigned(Source) And (Source Is TGLCapsule) Then
   Begin
      FParts := TGLCapsule(Source).FParts;
      FRadius := TGLCapsule(Source).FRadius;
   End;
   Inherited Assign(Source);
End;

Function TGLCapsule.AxisAlignedDimensionsUnscaled: TVector;
Var
   r, r1: Single;
Begin
   r := Abs(FRadius);
   r1 := Abs(FRadius);
   If r1 > r Then
      r := r1;
   Result := VectorMake(r, 0.5 * FHeight, r);
   // ScaleVector(Result, Scale.AsVector);
End;

Function TGLCapsule.RayCastIntersect(Const rayStart, rayVector: TVector; intersectPoint: PVector = nil; intersectNormal: PVector = nil): Boolean;
Const
   cOne: Single = 1;
Var
   locRayStart, locRayVector, ip: TVector;
   poly: Array [0 .. 2] Of Double;
   roots: TDoubleArray;
   minRoot: Double;
   t, tr2, invRayVector1, hTop, hBottom: Single;
   tPlaneMin, tPlaneMax: Single;
Begin
   Result := False;
   locRayStart := AbsoluteToLocal(rayStart);
   locRayVector := AbsoluteToLocal(rayVector);

   Case Alignment Of
      caTop:
      Begin
         hTop := 0;
         hBottom := -FHeight;
      End;
      caBottom:
      Begin
         hTop := FHeight;
         hBottom := 0;
      End;
      Else // caCenter
         hTop := FHeight * 0.5;
         hBottom := -hTop;
   End;

   If locRayVector.Y = 0 Then
   Begin
      // intersect if ray shot through the top/bottom planes
      If (locRayStart.X > hTop) Or (locRayStart.X < hBottom) Then
         Exit;
      tPlaneMin := -1E99;
      tPlaneMax := 1E99;
   End Else Begin
      invRayVector1 := cOne / locRayVector.Y;
      tr2 := Sqr(Radius);

      // compute intersection with topPlane
      t := (hTop - locRayStart.Y) * invRayVector1;
      If (t > 0) And (cyTop In Parts) Then
      Begin
         ip.X := locRayStart.X + t * locRayVector.X;
         ip.Z := locRayStart.Z + t * locRayVector.Z;
         If Sqr(ip.X) + Sqr(ip.Z) <= tr2 Then
         Begin
            // intersect with top plane
            If Assigned(intersectPoint) Then
               intersectPoint^ := LocalToAbsolute(VectorMake(ip.X, hTop, ip.Z, 1));
            If Assigned(intersectNormal) Then
               intersectNormal^ := LocalToAbsolute(YHmgVector);
            Result := True;
         End;
      End;
      tPlaneMin := t;
      tPlaneMax := t;
      // compute intersection with bottomPlane
      t := (hBottom - locRayStart.Y) * invRayVector1;
      If (t > 0) And (cyBottom In Parts) Then
      Begin
         ip.X := locRayStart.X + t * locRayVector.X;
         ip.Z := locRayStart.Z + t * locRayVector.Z;
         If (t < tPlaneMin) Or (Not (cyTop In Parts)) Then
         Begin
            If Sqr(ip.X) + Sqr(ip.Z) <= tr2 Then
            Begin
               // intersect with top plane
               If Assigned(intersectPoint) Then
                  intersectPoint^ := LocalToAbsolute(VectorMake(ip.X, hBottom, ip.Z, 1));
               If Assigned(intersectNormal) Then
                  intersectNormal^ := LocalToAbsolute(VectorNegate(YHmgVector));
               Result := True;
            End;
         End;
      End;
      If t < tPlaneMin Then
         tPlaneMin := t;
      If t > tPlaneMax Then
         tPlaneMax := t;
   End;
   If cySides In Parts Then
   Begin
      // intersect against cylinder infinite cylinder
      poly[0] := Sqr(locRayStart.X) + Sqr(locRayStart.Z) - Sqr(Radius);
      poly[1] := 2 * (locRayStart.X * locRayVector.X + locRayStart.Z * locRayVector.Z);
      poly[2] := Sqr(locRayVector.X) + Sqr(locRayVector.Z);
      roots := SolveQuadric(@poly);
      If MinPositiveCoef(roots, minRoot) Then
      Begin
         t := minRoot;
         If (t >= tPlaneMin) And (t < tPlaneMax) Then
         Begin
            If Assigned(intersectPoint) Or Assigned(intersectNormal) Then
            Begin
               ip := VectorCombine(locRayStart, locRayVector, 1, t);
               If Assigned(intersectPoint) Then
                  intersectPoint^ := LocalToAbsolute(ip);
               If Assigned(intersectNormal) Then
               Begin
                  ip.Y := 0;
                  ip.W := 0;
                  intersectNormal^ := LocalToAbsolute(ip);
               End;
            End;
            Result := True;
         End;
      End;
   End Else
      SetLength(roots, 0);
End;

Procedure TGLCapsule.Align(Const startPoint, endPoint: TVector);
Var
   dir: TAffineVector;
Begin
   AbsolutePosition := startPoint;
   VectorSubtract(endPoint, startPoint, dir);
   If Parent <> nil Then
      dir := Parent.AbsoluteToLocal(dir);
   Up.AsAffineVector := dir;
   FHeight := VectorLength(dir);
   Lift(FHeight * 0.5);
   Alignment := caCenter;
End;

Procedure TGLCapsule.Align(Const startObj, endObj: TGLBaseSceneObject);
Begin
   Align(startObj.AbsolutePosition, endObj.AbsolutePosition);
End;


Procedure TGLCapsule.Align(Const startPoint, endPoint: TAffineVector);
Begin
   Align(PointMake(startPoint), PointMake(endPoint));
End;

// ------------------
// ------------------ TGLAnnulus ------------------
// ------------------

Constructor TGLAnnulus.Create(AOwner: TComponent);
Begin
   Inherited Create(AOwner);
   FBottomInnerRadius := 0.3;
   FTopInnerRadius := 0.3;
   FTopRadius := 0.5;
   FParts := [anInnerSides, anOuterSides, anBottom, anTop];
End;

Procedure TGLAnnulus.SetBottomInnerRadius(Const aValue: Single);
Begin
   If aValue <> FBottomInnerRadius Then
   Begin
      FBottomInnerRadius := aValue;
      StructureChanged;
   End;
End;

Procedure TGLAnnulus.SetTopRadius(Const aValue: Single);
Begin
   If aValue <> FTopRadius Then
   Begin
      FTopRadius := aValue;
      StructureChanged;
   End;
End;

Procedure TGLAnnulus.SetTopInnerRadius(Const aValue: Single);
Begin
   If aValue <> FTopInnerRadius Then
   Begin
      FTopInnerRadius := aValue;
      StructureChanged;
   End;
End;

Procedure TGLAnnulus.SetParts(aValue: TAnnulusParts);
Begin
   If aValue <> FParts Then
   Begin
      FParts := aValue;
      StructureChanged;
   End;
End;

Procedure TGLAnnulus.BuildList(Var rci: TGLRenderContextInfo);
Var
   quadric: PGLUquadricObj;
Begin
   gl.PushMatrix;
   quadric := gluNewQuadric;
   SetupQuadricParams(quadric);
   gl.Rotatef(-90, 1, 0, 0);
   gl.Translatef(0, 0, -FHeight * 0.5);
     if [anBottom, anInnerSides] * FParts <> [] then
  begin
    SetInvertedQuadricOrientation(quadric);



    if anInnerSides in FParts then
      gluCylinder(quadric, FBottomInnerRadius, FTopInnerRadius, FHeight,
        FSlices, FStacks);
    SetNormalQuadricOrientation(quadric);
  end;

  if anOuterSides in FParts then
    gluCylinder(quadric, FBottomRadius, FTopRadius, FHeight, FSlices, FStacks);

  if anTop in FParts then
  begin
    gl.PushMatrix;
    gl.Translatef(0, 0, FHeight);
    gluDisk(quadric, FTopInnerRadius, FTopRadius, FSlices, FLoops);
    gl.PopMatrix;
  end;
  if [anBottom, anInnerSides] * FParts <> [] then
  begin
    // swap quadric orientation because top of a disk is defined as outside
    SetInvertedQuadricOrientation(quadric);
    if anBottom in FParts then
      gluDisk(quadric, FBottomInnerRadius, FBottomRadius, FSlices, FLoops);
  end;
  gluDeleteQuadric(quadric);
  gl.PopMatrix;

End;

Procedure TGLAnnulus.Assign(Source: TPersistent);
Begin
   If Assigned(Source) And (Source Is TGLAnnulus) Then
   Begin
      FParts := TGLAnnulus(Source).FParts;
      FTopRadius := TGLAnnulus(Source).FTopRadius;
      FTopInnerRadius := TGLAnnulus(Source).FTopInnerRadius;
      FBottomRadius := TGLAnnulus(Source).FBottomRadius;
      FBottomInnerRadius := TGLAnnulus(Source).FBottomInnerRadius;
   End;
   Inherited Assign(Source);
End;

Function TGLAnnulus.AxisAlignedDimensionsUnscaled: TVector;
Var
   r, r1: Single;
Begin
   r := Abs(FBottomRadius);
   r1 := Abs(FTopRadius);
   If r1 > r Then
      r := r1;
   Result := VectorMake(r, 0.5 * FHeight, r);
End;

Function TGLAnnulus.RayCastIntersect(Const rayStart, rayVector: TVector; intersectPoint, intersectNormal: PVector): Boolean;
Const
   cOne: Single = 1;
Var
   locRayStart, locRayVector, ip: TVector;
   poly: Array [0 .. 2] Of Double;
   t, tr2, invRayVector1: Single;
   tPlaneMin, tPlaneMax: Single;
   tir2, d2: Single;
   Root: Double;
   roots, tmpRoots: TDoubleArray;
   FirstIntersected: Boolean;
   h1, h2, hTop, hBot: Single;
   Draw1, Draw2: Boolean;
Begin
   Result := False;
   FirstIntersected := False;
   SetLength(tmpRoots, 0);
   locRayStart := AbsoluteToLocal(rayStart);
   locRayVector := AbsoluteToLocal(rayVector);

   hTop := Height * 0.5;
   hBot := -hTop;
   If locRayVector.Y < 0 Then
   Begin // Sort the planes according to the direction of view
      h1 := hTop; // Height of the 1st plane
      h2 := hBot; // Height of the 2nd plane
      Draw1 := (anTop In Parts); // 1st "cap" Must be drawn?
      Draw2 := (anBottom In Parts);
   End Else Begin
      h1 := hBot;
      h2 := hTop;
      Draw1 := (anBottom In Parts);
      Draw2 := (anTop In Parts);
   End; // if

   If locRayVector.Y = 0 Then
   Begin
      // intersect if ray shot through the top/bottom planes
      If (locRayStart.X > hTop) Or (locRayStart.X < hBot) Then
         Exit;
      tPlaneMin := -1E99;
      tPlaneMax := 1E99;
   End Else Begin
      invRayVector1 := cOne / locRayVector.Y;
      tr2 := Sqr(TopRadius);
      tir2 := Sqr(TopInnerRadius);
      FirstIntersected := False;

      // compute intersection with first plane
      t := (h1 - locRayStart.Y) * invRayVector1;
      If (t > 0) And Draw1 Then
      Begin
         ip.X := locRayStart.X + t * locRayVector.X;
         ip.Z := locRayStart.Z + t * locRayVector.Z;
         d2 := Sqr(ip.X) + Sqr(ip.Z);
         If (d2 <= tr2) And (d2 >= tir2) Then
         Begin
            // intersect with top plane
            FirstIntersected := True;
            If Assigned(intersectPoint) Then
               intersectPoint^ := LocalToAbsolute(VectorMake(ip.X, h1, ip.Z, 1));
            If Assigned(intersectNormal) Then
               intersectNormal^ := LocalToAbsolute(YHmgVector);
            Result := True;
         End;
      End;
      tPlaneMin := t;
      tPlaneMax := t;

      // compute intersection with second plane
      t := (h2 - locRayStart.Y) * invRayVector1;
      If (t > 0) And Draw2 Then
      Begin
         ip.X := locRayStart.X + t * locRayVector.X;
         ip.Z := locRayStart.Z + t * locRayVector.Z;
         d2 := Sqr(ip.X) + Sqr(ip.Z);
         If (t < tPlaneMin) Or (Not FirstIntersected) Then
         Begin
            If (d2 <= tr2) And (d2 >= tir2) Then
            Begin
               // intersect with top plane
               If Assigned(intersectPoint) Then
                  intersectPoint^ := LocalToAbsolute(VectorMake(ip.X, h2, ip.Z, 1));
               If Assigned(intersectNormal) Then
                  intersectNormal^ := LocalToAbsolute(VectorNegate(YHmgVector));
               Result := True;
            End;
         End;
      End;
      If t < tPlaneMin Then
      Begin
         tPlaneMin := t;
      End; // if
      If t > tPlaneMax Then
         tPlaneMax := t;
   End;

   Try
      SetLength(roots, 4);
      roots[0] := -1;
      roots[1] := -1;
      roots[2] := -1;
      roots[3] := -1; // By default, side is behind rayStart

      { Compute roots for outer cylinder }
      If anOuterSides In Parts Then
      Begin
         // intersect against infinite cylinder, will be cut by tPlaneMine and tPlaneMax
         poly[0] := Sqr(locRayStart.X) + Sqr(locRayStart.Z) - Sqr(TopRadius);
         poly[1] := 2 * (locRayStart.X * locRayVector.X + locRayStart.Z * locRayVector.Z);
         poly[2] := Sqr(locRayVector.X) + Sqr(locRayVector.Z);
         tmpRoots := SolveQuadric(@poly);
         // Intersect coordinates on rayVector (rayStart=0)
         If (High(tmpRoots) >= 0) And // Does root exist?
            ((tmpRoots[0] > tPlaneMin) And Not FirstIntersected) And
            // In the annulus and not masked by first cap
            ((tmpRoots[0] < tPlaneMax)) { // In the annulus } Then
            roots[0] := tmpRoots[0];
         If (High(tmpRoots) >= 1) And ((tmpRoots[1] > tPlaneMin) And Not FirstIntersected) And ((tmpRoots[1] < tPlaneMax)) Then
            roots[1] := tmpRoots[1];
      End; // if

      { Compute roots for inner cylinder }
      If anInnerSides In Parts Then
      Begin
         // intersect against infinite cylinder
         poly[0] := Sqr(locRayStart.X) + Sqr(locRayStart.Z) - Sqr(TopInnerRadius);
         poly[1] := 2 * (locRayStart.X * locRayVector.X + locRayStart.Z * locRayVector.Z);
         poly[2] := Sqr(locRayVector.X) + Sqr(locRayVector.Z);
         tmpRoots := SolveQuadric(@poly);
         If (High(tmpRoots) >= 0) And ((tmpRoots[0] > tPlaneMin) And Not FirstIntersected) And ((tmpRoots[0] < tPlaneMax)) Then
            roots[2] := tmpRoots[0];
         If (High(tmpRoots) >= 1) And ((tmpRoots[1] > tPlaneMin) And Not FirstIntersected) And ((tmpRoots[1] < tPlaneMax)) Then
            roots[3] := tmpRoots[1];
      End; // if

      { Find the first intersection point and compute its coordinates and normal }
      If MinPositiveCoef(roots, Root) Then
      Begin
         t := Root;
         If (t >= tPlaneMin) And (t < tPlaneMax) Then
         Begin
            If Assigned(intersectPoint) Or Assigned(intersectNormal) Then
            Begin
               ip := VectorCombine(locRayStart, locRayVector, 1, t);
               If Assigned(intersectPoint) Then
                  intersectPoint^ := LocalToAbsolute(ip);
               If Assigned(intersectNormal) Then
               Begin
                  ip.Y := 0;
                  ip.W := 0;
                  intersectNormal^ := LocalToAbsolute(ip);
               End;
            End;
            Result := True;
         End;
      End;

   Finally
      roots := nil;
      tmpRoots := nil;
   End; // finally
End;

// ------------------
// ------------------ TGLTorus ------------------
// ------------------

Constructor TGLTorus.Create(AOwner: TComponent);
Begin
   Inherited Create(AOwner);
   FRings := 25;
   FSides := 15;
   FMinorRadius := 0.1;
   FMajorRadius := 0.4;
   FStartAngle := 0.0;
   FStopAngle := 360.0;
   FParts := [toSides, toStartDisk, toStopDisk];
End;

Procedure TGLTorus.BuildList(Var rci: TGLRenderContextInfo);

   Procedure EmitVertex(ptr: PVertexRec; L1, L2: Integer);
   Begin
      XGL.TexCoord2fv(@ptr^.TexCoord);
      With GL Do
      Begin
         Normal3fv(@ptr^.Normal);
         If L1 > -1 Then
            VertexAttrib3fv(L1, @ptr.Tangent);
         If L2 > -1 Then
            VertexAttrib3fv(L2, @ptr.Binormal);
         Vertex3fv(@ptr^.Position);
      End;
   End;

Var
   i, j: Integer;
   Theta, Phi, Theta1, cosPhi, sinPhi, dist: Single;
   cosTheta1, sinTheta1: Single;
   ringDelta, sideDelta: Single;
   ringDir: TAffineVector;
   iFact, jFact: Single;
   pVertex: PVertexRec;
   TanLoc, BinLoc: Integer;
   MeshSize: Integer;
   MeshIndex: Integer;
   Vertex: TVertexRec;
Begin
   If FMesh = nil Then
   Begin
      MeshSize := 0;
      MeshIndex := 0;
      If toStartDisk In FParts Then
         MeshSize := MeshSize + 1;
      If toStopDisk In FParts Then
         MeshSize := MeshSize + 1;
      If toSides In FParts Then
         MeshSize := MeshSize + Integer(FRings) + 1;
      SetLength(FMesh, MeshSize);
      // handle texture generation
      ringDelta := ((FStopAngle - FStartAngle) / 360) * c2PI / FRings;
      sideDelta := c2PI / FSides;

      iFact := 1 / FRings;
      jFact := 1 / FSides;
      If toSides In FParts Then
      Begin
         Theta := DegToRadian(FStartAngle) - ringDelta;
         For i := FRings Downto 0 Do
         Begin
            SetLength(FMesh[i], FSides + 1);
            Theta1 := Theta + ringDelta;
            SinCosine(Theta1, sinTheta1, cosTheta1);
            Phi := 0;
            For j := FSides Downto 0 Do
            Begin
               Phi := Phi + sideDelta;
               SinCosine(Phi, sinPhi, cosPhi);
               dist := FMajorRadius + FMinorRadius * cosPhi;

               FMesh[i][j].Position := Vector3fMake(cosTheta1 * dist, -sinTheta1 * dist, FMinorRadius * sinPhi);
               ringDir := FMesh[i][j].Position;
               ringDir.Z := 0.0;
               NormalizeVector(ringDir);
               FMesh[i][j].Normal := Vector3fMake(cosTheta1 * cosPhi, -sinTheta1 * cosPhi, sinPhi);
               FMesh[i][j].Tangent := VectorCrossProduct(ZVector, ringDir);
               FMesh[i][j].Binormal := VectorCrossProduct(FMesh[i][j].Normal, FMesh[i][j].Tangent);
               FMesh[i][j].TexCoord := Vector2fMake(i * iFact, j * jFact);
            End;
            Theta := Theta1;
         End;
         MeshIndex := FRings + 1;
      End;

      If toStartDisk In FParts Then
      Begin
         SetLength(FMesh[MeshIndex], FSides + 1);
         Theta1 := DegToRadian(FStartAngle);
         SinCosine(Theta1, sinTheta1, cosTheta1);

         If toSides In FParts Then
         Begin
            For j := FSides Downto 0 Do
            Begin
               FMesh[MeshIndex][j].Position := FMesh[MeshIndex - 1][j].Position;
               FMesh[MeshIndex][j].Normal := FMesh[MeshIndex - 1][j].Tangent;
               FMesh[MeshIndex][j].Tangent := FMesh[MeshIndex - 1][j].Position;
               FMesh[MeshIndex][j].Tangent.Z := 0;
               FMesh[MeshIndex][j].Binormal := ZVector;
               FMesh[MeshIndex][j].TexCoord := FMesh[MeshIndex - 1][j].TexCoord;
               FMesh[MeshIndex][j].TexCoord.X := 0;
            End;
         End Else Begin
            Phi := 0;
            For j := FSides Downto 0 Do
            Begin
               Phi := Phi + sideDelta;
               SinCosine(Phi, sinPhi, cosPhi);
               dist := FMajorRadius + FMinorRadius * cosPhi;
               FMesh[MeshIndex][j].Position := Vector3fMake(cosTheta1 * dist, -sinTheta1 * dist, FMinorRadius * sinPhi);
               ringDir := FMesh[MeshIndex][j].Position;
               ringDir.Z := 0.0;
               NormalizeVector(ringDir);
               FMesh[MeshIndex][j].Normal := VectorCrossProduct(ZVector, ringDir);
               FMesh[MeshIndex][j].Tangent := ringDir;
               FMesh[MeshIndex][j].Binormal := ZVector;
               FMesh[MeshIndex][j].TexCoord := Vector2fMake(0, j * jFact);
            End;
         End;
         Vertex.Position := Vector3fMake(cosTheta1 * FMajorRadius, -sinTheta1 * FMajorRadius, 0);
         Vertex.Normal := FMesh[MeshIndex][0].Normal;
         Vertex.Tangent := FMesh[MeshIndex][0].Tangent;
         Vertex.Binormal := FMesh[MeshIndex][0].Binormal;
         Vertex.TexCoord := Vector2fMake(1, 1);
         MeshIndex := MeshIndex + 1;
      End;

      If toStopDisk In FParts Then
      Begin
         SetLength(FMesh[MeshIndex], FSides + 1);
         Theta1 := DegToRadian(FStopAngle);
         SinCosine(Theta1, sinTheta1, cosTheta1);

         If toSides In FParts Then
         Begin
            For j := FSides Downto 0 Do
            Begin
               FMesh[MeshIndex][j].Position := FMesh[0][j].Position;
               FMesh[MeshIndex][j].Normal := VectorNegate(FMesh[0][j].Tangent);
               FMesh[MeshIndex][j].Tangent := FMesh[0][j].Position;
               FMesh[MeshIndex][j].Tangent.Z := 0;
               FMesh[MeshIndex][j].Binormal := VectorNegate(ZVector);
               FMesh[MeshIndex][j].TexCoord := FMesh[0][j].TexCoord;
               FMesh[MeshIndex][j].TexCoord.X := 1;
            End;
         End Else Begin
            Phi := 0;
            For j := FSides Downto 0 Do
            Begin
               Phi := Phi + sideDelta;
               SinCosine(Phi, sinPhi, cosPhi);
               dist := FMajorRadius + FMinorRadius * cosPhi;
               FMesh[MeshIndex][j].Position := Vector3fMake(cosTheta1 * dist, -sinTheta1 * dist, FMinorRadius * sinPhi);
               ringDir := FMesh[MeshIndex][j].Position;
               ringDir.Z := 0.0;
               NormalizeVector(ringDir);
               FMesh[MeshIndex][j].Normal := VectorCrossProduct(ringDir, ZVector);
               FMesh[MeshIndex][j].Tangent := ringDir;
               FMesh[MeshIndex][j].Binormal := VectorNegate(ZVector);
               FMesh[MeshIndex][j].TexCoord := Vector2fMake(1, j * jFact);
            End;
         End;
         Vertex.Position := Vector3fMake(cosTheta1 * FMajorRadius, -sinTheta1 * FMajorRadius, 0);
         Vertex.Normal := FMesh[MeshIndex][0].Normal;
         Vertex.Tangent := FMesh[MeshIndex][0].Tangent;
         Vertex.Binormal := FMesh[MeshIndex][0].Binormal;
         Vertex.TexCoord := Vector2fMake(0, 0);
      End;
   End;
   Begin
      If GL.ARB_shader_objects And (rci.GLStates.CurrentProgram > 0) Then
      Begin
         TanLoc := gl.GetAttribLocation(rci.GLStates.CurrentProgram, Pansichar(TangentAttributeName));
         BinLoc := gl.GetAttribLocation(rci.GLStates.CurrentProgram, Pansichar(BinormalAttributeName));
      End Else Begin
         TanLoc := -1;
         BinLoc := TanLoc;
      End;

      MeshIndex := 0;

      If toSides In FParts Then
      Begin
         gl.Begin_(GL_TRIANGLES);
         For i := FRings - 1 Downto 0 Do
            For j := FSides - 1 Downto 0 Do
            Begin
               pVertex := @FMesh[i][j];
               EmitVertex(pVertex, TanLoc, BinLoc);

               pVertex := @FMesh[i][j + 1];
               EmitVertex(pVertex, TanLoc, BinLoc);

               pVertex := @FMesh[i + 1][j];
               EmitVertex(pVertex, TanLoc, BinLoc);

               pVertex := @FMesh[i + 1][j + 1];
               EmitVertex(pVertex, TanLoc, BinLoc);

               pVertex := @FMesh[i + 1][j];
               EmitVertex(pVertex, TanLoc, BinLoc);

               pVertex := @FMesh[i][j + 1];
               EmitVertex(pVertex, TanLoc, BinLoc);
            End;
         gl.End_;
         MeshIndex := FRings + 1;
      End;

      If toStartDisk In FParts Then
      Begin
         gl.Begin_(GL_TRIANGLE_FAN);
         pVertex := @Vertex;
         EmitVertex(pVertex, TanLoc, BinLoc);
         For j := 0 To FSides Do
         Begin
            pVertex := @FMesh[MeshIndex][j];
            EmitVertex(pVertex, TanLoc, BinLoc);
         End;
         gl.End_;
         MeshIndex := MeshIndex + 1;
      End;

      If toStopDisk In FParts Then
      Begin
         gl.Begin_(GL_TRIANGLE_FAN);
         pVertex := @Vertex;
         EmitVertex(pVertex, TanLoc, BinLoc);
         For j := FSides Downto 0 Do
         Begin
            pVertex := @FMesh[MeshIndex][j];
            EmitVertex(pVertex, TanLoc, BinLoc);
         End;
         gl.End_;
      End;
   End;
End;

Procedure TGLTorus.SetMajorRadius(Const aValue: Single);
Begin
   If FMajorRadius <> aValue Then
   Begin
      FMajorRadius := aValue;
      FMesh := nil;
      StructureChanged;
   End;
End;

Procedure TGLTorus.SetMinorRadius(Const aValue: Single);
Begin
   If FMinorRadius <> aValue Then
   Begin
      FMinorRadius := aValue;
      FMesh := nil;
      StructureChanged;
   End;
End;

Procedure TGLTorus.SetRings(aValue: Cardinal);
Begin
   If FRings <> aValue Then
   Begin
      FRings := aValue;
      If FRings < 2 Then
         FRings := 2;
      FMesh := nil;
      StructureChanged;
   End;
End;

Procedure TGLTorus.SetSides(aValue: Cardinal);
Begin
   If FSides <> aValue Then
   Begin
      FSides := aValue;
      If FSides < 3 Then
         FSides := 3;
      FMesh := nil;
      StructureChanged;
   End;
End;

Procedure TGLTorus.SetStartAngle(Const aValue: Single);
Begin
   If FStartAngle <> aValue Then
   Begin
      FStartAngle := aValue;
      FMesh := nil;
      StructureChanged;
   End;
End;

Procedure TGLTorus.SetStopAngle(Const aValue: Single);
Begin
   If FStopAngle <> aValue Then
   Begin
      FStopAngle := aValue;
      FMesh := nil;
      StructureChanged;
   End;
End;

Procedure TGLTorus.SetParts(aValue: TTorusParts);
Begin
   If aValue <> FParts Then
   Begin
      FParts := aValue;
      StructureChanged;
   End;
End;

Function TGLTorus.AxisAlignedDimensionsUnscaled: TVector;
Var
   r, r1: Single;
Begin
   r := Abs(FMajorRadius);
   r1 := Abs(FMinorRadius);
   Result := VectorMake(r + r1, r + r1, r1); // Danb
End;

Function TGLTorus.RayCastIntersect(Const rayStart, rayVector: TVector; intersectPoint: PVector = nil; intersectNormal: PVector = nil): Boolean;
Var
   i: Integer;
   fRo2, fRi2, fDE, fVal, r, nearest: Double;
   polynom: Array [0 .. 4] Of Double;
   polyRoots: TDoubleArray;
   localStart, localVector: TVector;
   vi, vc: TVector;
Begin
   // compute coefficients of quartic polynomial
   fRo2 := Sqr(MajorRadius);
   fRi2 := Sqr(MinorRadius);
   localStart := AbsoluteToLocal(rayStart);
   localVector := AbsoluteToLocal(rayVector);
   NormalizeVector(localVector);
   fDE := VectorDotProduct(localStart, localVector);
   fVal := VectorNorm(localStart) - (fRo2 + fRi2);

   polynom[0] := Sqr(fVal) - 4.0 * fRo2 * (fRi2 - Sqr(localStart.Z));
   polynom[1] := 4.0 * fDE * fVal + 8.0 * fRo2 * localVector.Z * localStart.Z;
   polynom[2] := 2.0 * fVal + 4.0 * Sqr(fDE) + 4.0 * fRo2 * Sqr(localVector.Z);
   polynom[3] := 4.0 * fDE;
   polynom[4] := 1;

   // solve the quartic
   polyRoots := SolveQuartic(@polynom[0]);

   // search for closest point
   Result := (Length(polyRoots) > 0);
   If Result Then
   Begin
      nearest := 1E20;
      For i := 0 To High(polyRoots) Do
      Begin
         r := polyRoots[i];
         If (r > 0) And (r < nearest) Then
         Begin
            nearest := r;
            Result := True;
         End;
      End;
      vi := VectorCombine(localStart, localVector, 1, nearest);
      If Assigned(intersectPoint) Then
         SetVector(intersectPoint^, LocalToAbsolute(vi));
      If Assigned(intersectNormal) Then
      Begin
         // project vi on local torus plane
         vc.X := vi.X;
         vc.Y := vi.Y;
         vc.Z := 0;
         // project vc on MajorRadius circle
         ScaleVector(vc, MajorRadius / (VectorLength(vc) + 0.000001));
         // calculate circle to intersect vector (gives normal);
         SubtractVector(vi, vc);
         // return to absolute coordinates and normalize
         vi.W := 0;
         SetVector(intersectNormal^, LocalToAbsolute(vi));
      End;
   End;
End;

// ------------------
// ------------------ TGLArrowLine ------------------
// ------------------

Constructor TGLArrowLine.Create(AOwner: TComponent);
Begin
   Inherited;
   FTopRadius := 0.1;
   BottomRadius := 0.1;
   fTopArrowHeadRadius := 0.2;
   fTopArrowHeadHeight := 0.5;
   fBottomArrowHeadRadius := 0.2;
   fBottomArrowHeadHeight := 0.5;
   FHeadStackingStyle := ahssStacked;
  { by default there is not much point having the top of the line (cylinder)
    showing as it is coincidental with the Toparrowhead bottom.
    Note I've defaulted to "vector" type arrows (arrow head on top only }
   FParts := [alLine, alTopArrow];
End;

Procedure TGLArrowLine.SetTopRadius(Const aValue: Single);
Begin
   If aValue <> FTopRadius Then
   Begin
      FTopRadius := aValue;
      StructureChanged;
   End;
End;

Procedure TGLArrowLine.SetTopArrowHeadHeight(Const aValue: Single);
Begin
   If aValue <> fTopArrowHeadHeight Then
   Begin
      fTopArrowHeadHeight := aValue;
      StructureChanged;
   End;
End;

Procedure TGLArrowLine.SetTopArrowHeadRadius(Const aValue: Single);
Begin
   If aValue <> fTopArrowHeadRadius Then
   Begin
      fTopArrowHeadRadius := aValue;
      StructureChanged;
   End;
End;

Procedure TGLArrowLine.SetBottomArrowHeadHeight(Const aValue: Single);
Begin
   If aValue <> fBottomArrowHeadHeight Then
   Begin
      fBottomArrowHeadHeight := aValue;
      StructureChanged;
   End;
End;

Procedure TGLArrowLine.SetBottomArrowHeadRadius(Const aValue: Single);
Begin
   If aValue <> fBottomArrowHeadRadius Then
   Begin
      fBottomArrowHeadRadius := aValue;
      StructureChanged;
   End;
End;

Procedure TGLArrowLine.SetParts(aValue: TArrowLineParts);
Begin
   If aValue <> FParts Then
   Begin
      FParts := aValue;
      StructureChanged;
   End;
End;

Procedure TGLArrowLine.SetHeadStackingStyle(Const val: TArrowHeadStackingStyle);
Begin
   If val <> FHeadStackingStyle Then
   Begin
      FHeadStackingStyle := val;
      StructureChanged;
   End;
End;

Procedure TGLArrowLine.BuildList(Var rci: TGLRenderContextInfo);
Var
   quadric: PGLUquadricObj;
   cylHeight, cylOffset, headInfluence: Single;
Begin
   Case HeadStackingStyle Of
      ahssCentered:
         headInfluence := 0.5;
      ahssIncluded:
         headInfluence := 1;
      Else // ahssStacked
         headInfluence := 0;
   End;
   cylHeight := Height;
   cylOffset := -FHeight * 0.5;
   // create a new quadric
   quadric := gluNewQuadric;
   SetupQuadricParams(quadric);
   // does the top arrow part - the cone
   If alTopArrow In Parts Then
   Begin
      cylHeight := cylHeight - TopArrowHeadHeight * headInfluence;
      gl.PushMatrix;
      gl.Translatef(0, 0, Height * 0.5 - TopArrowHeadHeight * headInfluence);
      gluCylinder(quadric, fTopArrowHeadRadius, 0, fTopArrowHeadHeight,
         Slices, Stacks);
      // top of a disk is defined as outside
      SetInvertedQuadricOrientation(quadric);
      If alLine In Parts Then
         gluDisk(quadric, FTopRadius, fTopArrowHeadRadius, Slices, FLoops)
      Else
         gluDisk(quadric, 0, fTopArrowHeadRadius, Slices, FLoops);
      gl.PopMatrix;
   End;
   // does the bottom arrow part - another cone
   If alBottomArrow In Parts Then
   Begin
      cylHeight := cylHeight - BottomArrowHeadHeight * headInfluence;
      cylOffset := cylOffset + BottomArrowHeadHeight * headInfluence;
      gl.PushMatrix;
      // make the bottom arrow point in the other direction
      gl.Rotatef(180, 1, 0, 0);
      gl.Translatef(0, 0, Height * 0.5 - BottomArrowHeadHeight * headInfluence);
      SetNormalQuadricOrientation(quadric);
      gluCylinder(quadric, fBottomArrowHeadRadius, 0, fBottomArrowHeadHeight,
         Slices, Stacks);
      // top of a disk is defined as outside
      SetInvertedQuadricOrientation(quadric);
      If alLine In Parts Then
         gluDisk(quadric, FBottomRadius, fBottomArrowHeadRadius, Slices, FLoops)
      Else
         gluDisk(quadric, 0, fBottomArrowHeadRadius, Slices, FLoops);
      gl.PopMatrix;
   End;
   // does the cylinder that makes the line
   If (cylHeight > 0) And (alLine In Parts) Then
   Begin
      gl.PushMatrix;
      gl.Translatef(0, 0, cylOffset);
      SetNormalQuadricOrientation(quadric);
      gluCylinder(quadric, FBottomRadius, FTopRadius, cylHeight, FSlices,
         FStacks);
      If Not (alTopArrow In Parts) Then
      Begin
         gl.PushMatrix;
         gl.Translatef(0, 0, cylHeight);
         gluDisk(quadric, 0, FTopRadius, FSlices, FLoops);
         gl.PopMatrix;
      End;
      If Not (alBottomArrow In Parts) Then
      Begin
         // swap quadric orientation because top of a disk is defined as outside
         SetInvertedQuadricOrientation(quadric);
         gluDisk(quadric, 0, FBottomRadius, FSlices, FLoops);
      End;
      gl.PopMatrix;
   End;
   gluDeleteQuadric(quadric);
End;

Procedure TGLArrowLine.Assign(Source: TPersistent);
Begin
   If Assigned(Source) And (Source Is TGLArrowLine) Then
   Begin
      FParts := TGLArrowLine(Source).FParts;
      FTopRadius := TGLArrowLine(Source).FTopRadius;
      fTopArrowHeadHeight := TGLArrowLine(Source).fTopArrowHeadHeight;
      fTopArrowHeadRadius := TGLArrowLine(Source).fTopArrowHeadRadius;
      fBottomArrowHeadHeight := TGLArrowLine(Source).fBottomArrowHeadHeight;
      fBottomArrowHeadRadius := TGLArrowLine(Source).fBottomArrowHeadRadius;
      FHeadStackingStyle := TGLArrowLine(Source).FHeadStackingStyle;
   End;
   Inherited Assign(Source);
End;

// ------------------
// ------------------ TGLArrowArc ------------------
// ------------------

Constructor TGLArrowArc.Create(AOwner: TComponent);
Begin
   Inherited;
   FStacks := 16;
   fArcRadius := 0.5;
   FStartAngle := 0;
   FStopAngle := 360;
   FTopRadius := 0.1;
   BottomRadius := 0.1;
   fTopArrowHeadRadius := 0.2;
   fTopArrowHeadHeight := 0.5;
   fBottomArrowHeadRadius := 0.2;
   fBottomArrowHeadHeight := 0.5;
   FHeadStackingStyle := ahssStacked;
   FParts := [aaArc, aaTopArrow];
End;

Procedure TGLArrowArc.SetArcRadius(Const aValue: Single);
Begin
   If fArcRadius <> aValue Then
   Begin
      fArcRadius := aValue;
      FMesh := nil;
      StructureChanged;
   End;
End;

Procedure TGLArrowArc.SetStartAngle(Const aValue: Single);
Begin
   If FStartAngle <> aValue Then
   Begin
      FStartAngle := aValue;
      FMesh := nil;
      StructureChanged;
   End;
End;

Procedure TGLArrowArc.SetStopAngle(Const aValue: Single);
Begin
   If FStopAngle <> aValue Then
   Begin
      FStopAngle := aValue;
      FMesh := nil;
      StructureChanged;
   End;
End;

Procedure TGLArrowArc.SetTopRadius(Const aValue: Single);
Begin
   If aValue <> FTopRadius Then
   Begin
      FTopRadius := aValue;
      FMesh := nil;
      StructureChanged;
   End;
End;

Procedure TGLArrowArc.SetTopArrowHeadHeight(Const aValue: Single);
Begin
   If aValue <> fTopArrowHeadHeight Then
   Begin
      fTopArrowHeadHeight := aValue;
      FMesh := nil;
      StructureChanged;
   End;
End;

Procedure TGLArrowArc.SetTopArrowHeadRadius(Const aValue: Single);
Begin
   If aValue <> fTopArrowHeadRadius Then
   Begin
      fTopArrowHeadRadius := aValue;
      FMesh := nil;
      StructureChanged;
   End;
End;

Procedure TGLArrowArc.SetBottomArrowHeadHeight(Const aValue: Single);
Begin
   If aValue <> fBottomArrowHeadHeight Then
   Begin
      fBottomArrowHeadHeight := aValue;
      FMesh := nil;
      StructureChanged;
   End;
End;

Procedure TGLArrowArc.SetBottomArrowHeadRadius(Const aValue: Single);
Begin
   If aValue <> fBottomArrowHeadRadius Then
   Begin
      fBottomArrowHeadRadius := aValue;
      FMesh := nil;
      StructureChanged;
   End;
End;

Procedure TGLArrowArc.SetParts(aValue: TArrowArcParts);
Begin
   If aValue <> FParts Then
   Begin
      FParts := aValue;
      FMesh := nil;
      StructureChanged;
   End;
End;

Procedure TGLArrowArc.SetHeadStackingStyle(Const val: TArrowHeadStackingStyle);
Begin
   If val <> FHeadStackingStyle Then
   Begin
      FHeadStackingStyle := val;
      FMesh := nil;
      StructureChanged;
   End;
End;

Procedure TGLArrowArc.BuildList(Var rci: TGLRenderContextInfo);

   Procedure EmitVertex(ptr: PVertexRec; L1, L2: Integer);
   Begin
      XGL.TexCoord2fv(@ptr^.TexCoord);
      gl.Normal3fv(@ptr^.Normal);
      If L1 > -1 Then
         gl.VertexAttrib3fv(L1, @ptr.Tangent);
      If L2 > -1 Then
         gl.VertexAttrib3fv(L2, @ptr.Binormal);
      gl.Vertex3fv(@ptr^.Position);
   End;

Var
   i, j: Integer;
   Theta, Phi, Theta1, cosPhi, sinPhi, dist: Single;
   cosTheta1, sinTheta1: Single;
   ringDelta, sideDelta: Single;
   ringDir: TAffineVector;
   iFact, jFact: Single;
   pVertex: PVertexRec;
   TanLoc, BinLoc: Integer;
   MeshSize: Integer;
   MeshIndex: Integer;
   ConeCenter: TVertexRec;
   StartOffset, StopOffset: Single;
Begin
   If FMesh = nil Then
   Begin
      MeshIndex := 0;
      MeshSize := 0;
      // Check Parts
      If aaArc In FParts Then
         MeshSize := MeshSize + FStacks + 1;
      If aaTopArrow In FParts Then
         MeshSize := MeshSize + 3
      Else
         MeshSize := MeshSize + 1;
      If aaBottomArrow In FParts Then
         MeshSize := MeshSize + 3
      Else
         MeshSize := MeshSize + 1;
      // Allocate Mesh
      SetLength(FMesh, MeshSize);

      Case FHeadStackingStyle Of
         ahssStacked:
         Begin
            StartOffset := 0;
            StopOffset := 0;
         End;
         ahssCentered:
         Begin
            If aaBottomArrow In Parts Then
               StartOffset :=
                  RadToDeg(ArcTan(0.5 * fBottomArrowHeadHeight / fArcRadius))
            Else
               StartOffset := 0;
            If aaTopArrow In Parts Then
               StopOffset :=
                  RadToDeg(ArcTan(0.5 * fTopArrowHeadHeight / fArcRadius))
            Else
               StopOffset := 0;
         End;
         ahssIncluded:
         Begin
            If aaBottomArrow In Parts Then
               StartOffset := RadToDeg(ArcTan(fBottomArrowHeadHeight / fArcRadius))
            Else
               StartOffset := 0;
            If aaTopArrow In Parts Then
               StopOffset := RadToDeg(ArcTan(fTopArrowHeadHeight / fArcRadius))
            Else
               StopOffset := 0;
         End;
         Else StartOffset := 0;
            StopOffset := 0;
      End;

      // handle texture generation
      ringDelta := (((FStopAngle - StopOffset) - (FStartAngle + StartOffset)) / 360) * c2PI / FStacks;
      sideDelta := c2PI / FSlices;

      iFact := 1 / FStacks;
      jFact := 1 / FSlices;
      If aaArc In FParts Then
      Begin
         Theta := DegToRadian(FStartAngle + StartOffset) - ringDelta;
         For i := FStacks Downto 0 Do
         Begin
            SetLength(FMesh[i], FSlices + 1);
            Theta1 := Theta + ringDelta;
            SinCosine(Theta1, sinTheta1, cosTheta1);
            Phi := 0;
            For j := FSlices Downto 0 Do
            Begin
               Phi := Phi + sideDelta;
               SinCosine(Phi, sinPhi, cosPhi);
               dist := fArcRadius + Lerp(FTopRadius, FBottomRadius, i * iFact) * cosPhi;

               FMesh[i][j].Position := Vector3fMake(cosTheta1 * dist, -sinTheta1 * dist, Lerp(FTopRadius, FBottomRadius, i * iFact) * sinPhi);
               ringDir := FMesh[i][j].Position;
               ringDir.Z := 0.0;
               NormalizeVector(ringDir);
               FMesh[i][j].Normal := Vector3fMake(cosTheta1 * cosPhi, -sinTheta1 * cosPhi, sinPhi);
               FMesh[i][j].Tangent := VectorCrossProduct(ZVector, ringDir);
               FMesh[i][j].Binormal := VectorCrossProduct(FMesh[i][j].Normal, FMesh[i][j].Tangent);
               FMesh[i][j].TexCoord := Vector2fMake(i * iFact, j * jFact);
            End;
            Theta := Theta1;
         End;
         MeshIndex := FStacks + 1;
         Begin
            If GL.ARB_shader_objects And (rci.GLStates.CurrentProgram > 0) Then
            Begin
               TanLoc := gl.GetAttribLocation(rci.GLStates.CurrentProgram, Pansichar(TangentAttributeName));
               BinLoc := gl.GetAttribLocation(rci.GLStates.CurrentProgram, Pansichar(BinormalAttributeName));
            End Else Begin
               TanLoc := -1;
               BinLoc := TanLoc;
            End;

            gl.Begin_(GL_TRIANGLES);
            For i := FStacks - 1 Downto 0 Do
               For j := FSlices - 1 Downto 0 Do
               Begin
                  pVertex := @FMesh[i][j];
                  EmitVertex(pVertex, TanLoc, BinLoc);

                  pVertex := @FMesh[i][j + 1];
                  EmitVertex(pVertex, TanLoc, BinLoc);

                  pVertex := @FMesh[i + 1][j];
                  EmitVertex(pVertex, TanLoc, BinLoc);

                  pVertex := @FMesh[i + 1][j + 1];
                  EmitVertex(pVertex, TanLoc, BinLoc);

                  pVertex := @FMesh[i + 1][j];
                  EmitVertex(pVertex, TanLoc, BinLoc);

                  pVertex := @FMesh[i][j + 1];
                  EmitVertex(pVertex, TanLoc, BinLoc);
               End;
            gl.End_;
         End;
      End;

      // Build Arrow or start cap
      If aaBottomArrow In FParts Then
      Begin
         SetLength(FMesh[MeshIndex], FSlices + 1);
         SetLength(FMesh[MeshIndex + 1], FSlices + 1);
         SetLength(FMesh[MeshIndex + 2], FSlices + 1);
         Theta1 := DegToRadian(FStartAngle + StartOffset);
         SinCosine(Theta1, sinTheta1, cosTheta1);

         ConeCenter.Position := Vector3fMake(cosTheta1 * fArcRadius, -sinTheta1 * fArcRadius, 0);

         Phi := 0;
         For j := FSlices Downto 0 Do
         Begin
            Phi := Phi + sideDelta;
            SinCosine(Phi, sinPhi, cosPhi);
            dist := fArcRadius + fBottomArrowHeadRadius * cosPhi;

            // Cap
            FMesh[MeshIndex][J].Position := Vector3fMake(cosTheta1 * dist, -sinTheta1 * dist, fBottomArrowHeadRadius * sinPhi);
            ringDir := FMesh[MeshIndex][j].Position;
            ringDir.Z := 0.0;
            NormalizeVector(ringDir);
            FMesh[MeshIndex][j].Normal := VectorCrossProduct(ringDir, ZVector);
            FMesh[MeshIndex][j].Tangent := ringDir;
            FMesh[MeshIndex][j].Binormal := ZVector;
            FMesh[MeshIndex][j].TexCoord := Vector2fMake(1, j * jFact);

            // Cone
            FMesh[MeshIndex + 1][j].Position := Vector3fMake(cosTheta1 * dist, -sinTheta1 * dist, fBottomArrowHeadRadius * sinPhi);
            FMesh[MeshIndex + 2][j].Position := VectorAdd(ConeCenter.Position, Vector3fMake(sinTheta1 * fBottomArrowHeadHeight, cosTheta1 * fBottomArrowHeadHeight, 0));

            FMesh[MeshIndex + 1][j].Tangent :=
               VectorNormalize(VectorSubtract(FMesh[MeshIndex + 1][j].Position, FMesh[MeshIndex + 2][j].Position));
            FMesh[MeshIndex + 2][j].Tangent := FMesh[MeshIndex + 1][j].Tangent;

            FMesh[MeshIndex + 1][j].Binormal := Vector3fMake(cosTheta1 * -sinPhi, sinTheta1 * sinPhi, cosPhi);
            FMesh[MeshIndex + 2][j].Binormal := FMesh[MeshIndex + 1][j].Binormal;

            FMesh[MeshIndex + 1][j].Normal :=
               VectorCrossProduct(FMesh[MeshIndex + 1][j].Binormal, FMesh[MeshIndex + 1][j].Tangent);
            FMesh[MeshIndex + 2][j].Normal := FMesh[MeshIndex + 1][j].Normal;

            FMesh[MeshIndex + 1][j].TexCoord := Vector2fMake(0, j * jFact);
            FMesh[MeshIndex + 2][j].TexCoord := Vector2fMake(1, j * jFact);
         End;

         ConeCenter.Normal := FMesh[MeshIndex][0].Normal;
         ConeCenter.Tangent := FMesh[MeshIndex][0].Tangent;
         ConeCenter.Binormal := FMesh[MeshIndex][0].Binormal;
         ConeCenter.TexCoord := Vector2fMake(0, 0);
         Begin
            If GL.ARB_shader_objects And (rci.GLStates.CurrentProgram > 0) Then
            Begin
               TanLoc := gl.GetAttribLocation(rci.GLStates.CurrentProgram, Pansichar(TangentAttributeName));
               BinLoc := gl.GetAttribLocation(rci.GLStates.CurrentProgram, Pansichar(BinormalAttributeName));
            End Else Begin
               TanLoc := -1;
               BinLoc := TanLoc;
            End;

            gl.Begin_(GL_TRIANGLE_FAN);
            pVertex := @ConeCenter;
            EmitVertex(pVertex, TanLoc, BinLoc);
            For j := FSlices Downto 0 Do
            Begin
               pVertex := @FMesh[MeshIndex][j];
               EmitVertex(pVertex, TanLoc, BinLoc);
            End;
            gl.End_;

            gl.Begin_(GL_TRIANGLES);

            For j := FSlices - 1 Downto 0 Do
            Begin
               pVertex := @FMesh[MeshIndex + 1][j];
               EmitVertex(pVertex, TanLoc, BinLoc);

               pVertex := @FMesh[MeshIndex + 1][j + 1];
               EmitVertex(pVertex, TanLoc, BinLoc);

               pVertex := @FMesh[MeshIndex + 2][j];
               EmitVertex(pVertex, TanLoc, BinLoc);

               pVertex := @FMesh[MeshIndex + 2][j + 1];
               EmitVertex(pVertex, TanLoc, BinLoc);

               pVertex := @FMesh[MeshIndex + 2][j];
               EmitVertex(pVertex, TanLoc, BinLoc);

               pVertex := @FMesh[MeshIndex + 1][j + 1];
               EmitVertex(pVertex, TanLoc, BinLoc);
            End;
            gl.End_;
         End;
         MeshIndex := MeshIndex + 3;
      End Else Begin
         SetLength(FMesh[MeshIndex], FSlices + 1);
         Theta1 := DegToRadian(FStartAngle);
         SinCosine(Theta1, sinTheta1, cosTheta1);

         Phi := 0;
         For j := FSlices Downto 0 Do
         Begin
            Phi := Phi + sideDelta;
            SinCosine(Phi, sinPhi, cosPhi);
            dist := fArcRadius + fBottomRadius * cosPhi;
            FMesh[MeshIndex][j].Position := Vector3fMake(cosTheta1 * dist, -sinTheta1 * dist, FBottomRadius * sinPhi);
            ringDir := FMesh[MeshIndex][j].Position;
            ringDir.Z := 0.0;
            NormalizeVector(ringDir);
            FMesh[MeshIndex][j].Normal := VectorCrossProduct(ZVector, ringDir);
            FMesh[MeshIndex][j].Tangent := ringDir;
            FMesh[MeshIndex][j].Binormal := ZVector;
            FMesh[MeshIndex][j].TexCoord := Vector2fMake(0, j * jFact);
         End;

         ConeCenter.Position := Vector3fMake(cosTheta1 * fArcRadius, -sinTheta1 * fArcRadius, 0);
         ConeCenter.Normal := FMesh[MeshIndex][0].Normal;
         ConeCenter.Tangent := FMesh[MeshIndex][0].Tangent;
         ConeCenter.Binormal := FMesh[MeshIndex][0].Binormal;
         ConeCenter.TexCoord := Vector2fMake(1, 1);
         With GL Do
         Begin
            If ARB_shader_objects And (rci.GLStates.CurrentProgram > 0) Then
            Begin
               TanLoc := GetAttribLocation(rci.GLStates.CurrentProgram, Pansichar(TangentAttributeName));
               BinLoc := GetAttribLocation(rci.GLStates.CurrentProgram, Pansichar(BinormalAttributeName));
            End Else Begin
               TanLoc := -1;
               BinLoc := TanLoc;
            End;
            Begin_(GL_TRIANGLE_FAN);
            pVertex := @ConeCenter;
            EmitVertex(pVertex, TanLoc, BinLoc);
            For j := 0 To FSlices Do
            Begin
               pVertex := @FMesh[MeshIndex][j];
               EmitVertex(pVertex, TanLoc, BinLoc);
            End;
            End_;
         End;
         MeshIndex := MeshIndex + 1;
      End;

      If aaTopArrow In FParts Then
      Begin
         SetLength(FMesh[MeshIndex], FSlices + 1);
         SetLength(FMesh[MeshIndex + 1], FSlices + 1);
         SetLength(FMesh[MeshIndex + 2], FSlices + 1);
         Theta1 := DegToRadian(FStopAngle - StopOffset);
         SinCosine(Theta1, sinTheta1, cosTheta1);

         ConeCenter.Position := Vector3fMake(cosTheta1 * fArcRadius, -sinTheta1 * fArcRadius, 0);

         Phi := 0;
         For j := FSlices Downto 0 Do
         Begin
            Phi := Phi + sideDelta;
            SinCosine(Phi, sinPhi, cosPhi);
            dist := fArcRadius + fTopArrowHeadRadius * cosPhi;

            // Cap
            FMesh[MeshIndex][j].Position := Vector3fMake(cosTheta1 * dist, -sinTheta1 * dist, fTopArrowHeadRadius * sinPhi);
            ringDir := FMesh[MeshIndex][j].Position;
            ringDir.Z := 0.0;
            NormalizeVector(ringDir);
            FMesh[MeshIndex][j].Normal := VectorCrossProduct(ZVector, ringDir);
            FMesh[MeshIndex][j].Tangent := ringDir;
            FMesh[MeshIndex][j].Binormal := ZVector;
            FMesh[MeshIndex][j].TexCoord := Vector2fMake(0, j * jFact);

            // Cone
            FMesh[MeshIndex + 1][j].Position := Vector3fMake(cosTheta1 * dist, -sinTheta1 * dist, fTopArrowHeadRadius * sinPhi);
            FMesh[MeshIndex + 2][j].Position := VectorSubtract(ConeCenter.Position, Vector3fMake(sinTheta1 * fTopArrowHeadHeight, cosTheta1 * fTopArrowHeadHeight, 0));

            FMesh[MeshIndex + 1][j].Tangent :=
               VectorNormalize(VectorSubtract(FMesh[MeshIndex + 2][j].Position, FMesh[MeshIndex + 1][j].Position));
            FMesh[MeshIndex + 2][j].Tangent := FMesh[MeshIndex + 1][j].Tangent;

            FMesh[MeshIndex + 1][j].Binormal := Vector3fMake(cosTheta1 * -sinPhi, sinTheta1 * sinPhi, cosPhi);
            FMesh[MeshIndex + 2][j].Binormal := FMesh[MeshIndex + 1][j].Binormal;

            FMesh[MeshIndex + 1][j].Normal :=
               VectorCrossProduct(FMesh[MeshIndex + 1][j].Binormal, FMesh[MeshIndex + 1][j].Tangent);
            FMesh[MeshIndex + 2][j].Normal := FMesh[MeshIndex + 1][j].Normal;

            FMesh[MeshIndex + 1][j].TexCoord := Vector2fMake(1, j * jFact);
            FMesh[MeshIndex + 2][j].TexCoord := Vector2fMake(0, j * jFact);
         End;

         ConeCenter.Normal := FMesh[MeshIndex][0].Normal;
         ConeCenter.Tangent := FMesh[MeshIndex][0].Tangent;
         ConeCenter.Binormal := FMesh[MeshIndex][0].Binormal;
         ConeCenter.TexCoord := Vector2fMake(1, 1);
         Begin
            If GL.ARB_shader_objects And (rci.GLStates.CurrentProgram > 0) Then
            Begin
               TanLoc := gl.GetAttribLocation(rci.GLStates.CurrentProgram, Pansichar(TangentAttributeName));
               BinLoc := gl.GetAttribLocation(rci.GLStates.CurrentProgram, Pansichar(BinormalAttributeName));
            End Else Begin
               TanLoc := -1;
               BinLoc := TanLoc;
            End;

            gl.Begin_(GL_TRIANGLE_FAN);
            pVertex := @ConeCenter;
            EmitVertex(pVertex, TanLoc, BinLoc);
            For j := 0 To FSlices Do
            Begin
               pVertex := @FMesh[MeshIndex][j];
               EmitVertex(pVertex, TanLoc, BinLoc);
            End;
            gl.End_;

            gl.Begin_(GL_TRIANGLES);

            For j := FSlices - 1 Downto 0 Do
            Begin
               pVertex := @FMesh[MeshIndex + 2][j];
               EmitVertex(pVertex, TanLoc, BinLoc);

               pVertex := @FMesh[MeshIndex + 2][j + 1];
               EmitVertex(pVertex, TanLoc, BinLoc);

               pVertex := @FMesh[MeshIndex + 1][j];
               EmitVertex(pVertex, TanLoc, BinLoc);

               pVertex := @FMesh[MeshIndex + 1][j + 1];
               EmitVertex(pVertex, TanLoc, BinLoc);

               pVertex := @FMesh[MeshIndex + 1][j];
               EmitVertex(pVertex, TanLoc, BinLoc);

               pVertex := @FMesh[MeshIndex + 2][j + 1];
               EmitVertex(pVertex, TanLoc, BinLoc);
            End;
            gl.End_;
         End;
      End Else Begin
         SetLength(FMesh[MeshIndex], FSlices + 1);
         Theta1 := DegToRadian(FStopAngle);
         SinCosine(Theta1, sinTheta1, cosTheta1);

         Phi := 0;
         For j := FSlices Downto 0 Do
         Begin
            Phi := Phi + sideDelta;
            SinCosine(Phi, sinPhi, cosPhi);
            dist := fArcRadius + fTopRadius * cosPhi;
            FMesh[MeshIndex][j].Position := Vector3fMake(cosTheta1 * dist, -sinTheta1 * dist, fTopRadius * sinPhi);
            ringDir := FMesh[MeshIndex][j].Position;
            ringDir.Z := 0.0;
            NormalizeVector(ringDir);
            FMesh[MeshIndex][j].Normal := VectorCrossProduct(ringDir, ZVector);
            FMesh[MeshIndex][j].Tangent := ringDir;
            FMesh[MeshIndex][j].Binormal := VectorNegate(ZVector);
            FMesh[MeshIndex][j].TexCoord := Vector2fMake(1, j * jFact);
         End;
         ConeCenter.Position := Vector3fMake(cosTheta1 * fArcRadius, -sinTheta1 * fArcRadius, 0);
         ConeCenter.Normal := FMesh[MeshIndex][0].Normal;
         ConeCenter.Tangent := FMesh[MeshIndex][0].Tangent;
         ConeCenter.Binormal := FMesh[MeshIndex][0].Binormal;
         ConeCenter.TexCoord := Vector2fMake(0, 0);
         Begin
            If GL.ARB_shader_objects And (rci.GLStates.CurrentProgram > 0) Then
            Begin
               TanLoc := gl.GetAttribLocation(rci.GLStates.CurrentProgram, Pansichar(TangentAttributeName));
               BinLoc := gl.GetAttribLocation(rci.GLStates.CurrentProgram, Pansichar(BinormalAttributeName));
            End Else Begin
               TanLoc := -1;
               BinLoc := TanLoc;
            End;
            gl.Begin_(GL_TRIANGLE_FAN);
            pVertex := @ConeCenter;
            EmitVertex(pVertex, TanLoc, BinLoc);
            For j := FSlices Downto 0 Do
            Begin
               pVertex := @FMesh[MeshIndex][j];
               EmitVertex(pVertex, TanLoc, BinLoc);
            End;
            gl.End_;
         End;
      End;
   End;
End;

Procedure TGLArrowArc.Assign(Source: TPersistent);
Begin
   If Assigned(Source) And (Source Is TGLArrowLine) Then
   Begin
      FStartAngle := TGLArrowArc(Source).FStartAngle;
      FStopAngle := TGLArrowArc(Source).FStopAngle;
      fArcRadius := TGLArrowArc(Source).fArcRadius;
      FParts := TGLArrowArc(Source).FParts;
      FTopRadius := TGLArrowArc(Source).FTopRadius;
      fTopArrowHeadHeight := TGLArrowArc(Source).fTopArrowHeadHeight;
      fTopArrowHeadRadius := TGLArrowArc(Source).fTopArrowHeadRadius;
      fBottomArrowHeadHeight := TGLArrowArc(Source).fBottomArrowHeadHeight;
      fBottomArrowHeadRadius := TGLArrowArc(Source).fBottomArrowHeadRadius;
      FHeadStackingStyle := TGLArrowArc(Source).FHeadStackingStyle;
   End;
   Inherited Assign(Source);
End;

// ------------------
// ------------------ TGLFrustrum ------------------
// ------------------

Constructor TGLFrustrum.Create(AOwner: TComponent);
Begin
   Inherited Create(AOwner);
   FApexHeight := 1;
   FBaseWidth := 1;
   FBaseDepth := 1;
   FHeight := 0.5;
   FParts := cAllFrustrumParts;
   FNormalDirection := ndOutside;
End;

Procedure TGLFrustrum.BuildList(Var rci: TGLRenderContextInfo);
Var
   HBW, HBD: Single; // half of width, half of depth at base
   HTW, HTD: Single; // half of width, half of depth at top of frustrum
   HFH: Single; // half of height, for align to center
   Sign: Single; // +1 or -1
   angle: Single; // in radians
   ASin, ACos: Single;
Begin
   If FNormalDirection = ndInside Then
      Sign := -1
   Else
      Sign := 1;
   HBW := FBaseWidth * 0.5;
   HBD := FBaseDepth * 0.5;
   HTW := HBW * (FApexHeight - FHeight) / FApexHeight;
   HTD := HBD * (FApexHeight - FHeight) / FApexHeight;
   HFH := FHeight * 0.5;

   gl.Begin_(GL_QUADS);

   If [fpFront, fpBack] * FParts <> [] Then
   Begin
      angle := ArcTan(FApexHeight / HBD);
      // angle of front plane with bottom plane
      SinCosine(angle, ASin, ACos);
      If fpFront In FParts Then
      Begin
         gl.Normal3f(0, Sign * ACos, Sign * ASin);
         XGL.TexCoord2fv(@XYTexPoint);
         gl.Vertex3f(HTW, HFH, HTD);
         XGL.TexCoord2fv(@YTexPoint);
         gl.Vertex3f(-HTW, HFH, HTD);
         XGL.TexCoord2fv(@NullTexPoint);
         gl.Vertex3f(-HBW, -HFH, HBD);
         XGL.TexCoord2fv(@XTexPoint);
         gl.Vertex3f(HBW, -HFH, HBD);
      End;
      If fpBack In FParts Then
      Begin
         gl.Normal3f(0, Sign * ACos, -Sign * ASin);
         XGL.TexCoord2fv(@YTexPoint);
         gl.Vertex3f(HTW, HFH, -HTD);
         XGL.TexCoord2fv(@NullTexPoint);
         gl.Vertex3f(HBW, -HFH, -HBD);
         XGL.TexCoord2fv(@XTexPoint);
         gl.Vertex3f(-HBW, -HFH, -HBD);
         XGL.TexCoord2fv(@XYTexPoint);
         gl.Vertex3f(-HTW, HFH, -HTD);
      End;
   End;

   If [fpLeft, fpRight] * FParts <> [] Then
   Begin
      angle := ArcTan(FApexHeight / HBW); // angle of side plane with bottom plane
      SinCosine(angle, ASin, ACos);
      If fpLeft In FParts Then
      Begin
         gl.Normal3f(-Sign * ASin, Sign * ACos, 0);
         XGL.TexCoord2fv(@XYTexPoint);
         gl.Vertex3f(-HTW, HFH, HTD);
         XGL.TexCoord2fv(@YTexPoint);
         gl.Vertex3f(-HTW, HFH, -HTD);
         XGL.TexCoord2fv(@NullTexPoint);
         gl.Vertex3f(-HBW, -HFH, -HBD);
         XGL.TexCoord2fv(@XTexPoint);
         gl.Vertex3f(-HBW, -HFH, HBD);
      End;
      If fpRight In FParts Then
      Begin
         gl.Normal3f(Sign * ASin, Sign * ACos, 0);
         XGL.TexCoord2fv(@YTexPoint);
         gl.Vertex3f(HTW, HFH, HTD);
         XGL.TexCoord2fv(@NullTexPoint);
         gl.Vertex3f(HBW, -HFH, HBD);
         XGL.TexCoord2fv(@XTexPoint);
         gl.Vertex3f(HBW, -HFH, -HBD);
         XGL.TexCoord2fv(@XYTexPoint);
         gl.Vertex3f(HTW, HFH, -HTD);
      End;
   End;

   If (fpTop In FParts) And (FHeight < FApexHeight) Then
   Begin
      gl.Normal3f(0, Sign, 0);
      XGL.TexCoord2fv(@YTexPoint);
      gl.Vertex3f(-HTW, HFH, -HTD);
      XGL.TexCoord2fv(@NullTexPoint);
      gl.Vertex3f(-HTW, HFH, HTD);
      XGL.TexCoord2fv(@XTexPoint);
      gl.Vertex3f(HTW, HFH, HTD);
      XGL.TexCoord2fv(@XYTexPoint);
      gl.Vertex3f(HTW, HFH, -HTD);
   End;
   If fpBottom In FParts Then
   Begin
      gl.Normal3f(0, -Sign, 0);
      XGL.TexCoord2fv(@NullTexPoint);
      gl.Vertex3f(-HBW, -HFH, -HBD);
      XGL.TexCoord2fv(@XTexPoint);
      gl.Vertex3f(HBW, -HFH, -HBD);
      XGL.TexCoord2fv(@XYTexPoint);
      gl.Vertex3f(HBW, -HFH, HBD);
      XGL.TexCoord2fv(@YTexPoint);
      gl.Vertex3f(-HBW, -HFH, HBD);
   End;

   gl.End_;
End;

Procedure TGLFrustrum.SetApexHeight(Const aValue: Single);
Begin
   If (aValue <> FApexHeight) And (aValue >= 0) Then
   Begin
      FApexHeight := aValue;
      If FHeight > aValue Then
         FHeight := aValue;
      StructureChanged;
   End;
End;

Procedure TGLFrustrum.SetBaseDepth(Const aValue: Single);
Begin
   If (aValue <> FBaseDepth) And (aValue >= 0) Then
   Begin
      FBaseDepth := aValue;
      StructureChanged;
   End;
End;

Procedure TGLFrustrum.SetBaseWidth(Const aValue: Single);
Begin
   If (aValue <> FBaseWidth) And (aValue >= 0) Then
   Begin
      FBaseWidth := aValue;
      StructureChanged;
   End;
End;

Procedure TGLFrustrum.SetHeight(Const aValue: Single);
Begin
   If (aValue <> FHeight) And (aValue >= 0) Then
   Begin
      FHeight := aValue;
      If FApexHeight < aValue Then
         FApexHeight := aValue;
      StructureChanged;
   End;
End;

Procedure TGLFrustrum.SetParts(aValue: TFrustrumParts);
Begin
   If aValue <> FParts Then
   Begin
      FParts := aValue;
      StructureChanged;
   End;
End;

Procedure TGLFrustrum.SetNormalDirection(aValue: TNormalDirection);
Begin
   If aValue <> FNormalDirection Then
   Begin
      FNormalDirection := aValue;
      StructureChanged;
   End;
End;

Procedure TGLFrustrum.Assign(Source: TPersistent);
Begin
   If Assigned(Source) And (Source Is TGLFrustrum) Then
   Begin
      FApexHeight := TGLFrustrum(Source).FApexHeight;
      FBaseDepth := TGLFrustrum(Source).FBaseDepth;
      FBaseWidth := TGLFrustrum(Source).FBaseWidth;
      FHeight := TGLFrustrum(Source).FHeight;
      FParts := TGLFrustrum(Source).FParts;
      FNormalDirection := TGLFrustrum(Source).FNormalDirection;
   End;
   Inherited Assign(Source);
End;

Function TGLFrustrum.TopDepth: Single;
Begin
   Result := FBaseDepth * (FApexHeight - FHeight) / FApexHeight;
End;

Function TGLFrustrum.TopWidth: Single;
Begin
   Result := FBaseWidth * (FApexHeight - FHeight) / FApexHeight;
End;

Procedure TGLFrustrum.DefineProperties(Filer: TFiler);
Begin
   Inherited;
   Filer.DefineBinaryProperty('FrustrumSize', ReadData, WriteData,
      (FApexHeight <> 1) Or (FBaseDepth <> 1) Or (FBaseWidth <> 1) Or (FHeight <> 0.5));
End;

Procedure TGLFrustrum.ReadData(Stream: TStream);
Begin
   With Stream Do
   Begin
      Read(FApexHeight, SizeOf(FApexHeight));
      Read(FBaseDepth, SizeOf(FBaseDepth));
      Read(FBaseWidth, SizeOf(FBaseWidth));
      Read(FHeight, SizeOf(FHeight));
   End;
End;

Procedure TGLFrustrum.WriteData(Stream: TStream);
Begin
   With Stream Do
   Begin
      Write(FApexHeight, SizeOf(FApexHeight));
      Write(FBaseDepth, SizeOf(FBaseDepth));
      Write(FBaseWidth, SizeOf(FBaseWidth));
      Write(FHeight, SizeOf(FHeight));
   End;
End;

Function TGLFrustrum.AxisAlignedBoundingBoxUnscaled: TAABB;
Var
   aabb: TAABB;
   child: TGLBaseSceneObject;
   i: Integer;
Begin
   SetAABB(Result, AxisAlignedDimensionsUnscaled);
   OffsetAABB(Result, VectorMake(0, FHeight * 0.5, 0));

   // not tested for child objects
   For i := 0 To Count - 1 Do
   Begin
      child := TGLBaseSceneObject(Children[i]);
      aabb := child.AxisAlignedBoundingBoxUnscaled;
      AABBTransform(aabb, child.Matrix^);
      AddAABB(Result, aabb);
   End;
End;

Function TGLFrustrum.AxisAlignedDimensionsUnscaled: TVector;
Begin
   Result.X := FBaseWidth * 0.5;
   Result.Y := FHeight * 0.5;
   Result.Z := FBaseDepth * 0.5;
   Result.W := 0;
End;

// ------------------
// ------------------ TGLPolygon ------------------
// ------------------

Constructor TGLPolygon.Create(AOwner: TComponent);
Begin
   Inherited Create(AOwner);
   FParts := [ppTop, ppBottom];
End;


Destructor TGLPolygon.Destroy;
Begin
   Inherited Destroy;
End;

Procedure TGLPolygon.SetParts(Const val: TPolygonParts);
Begin
   If FParts <> val Then
   Begin
      FParts := val;
      StructureChanged;
   End;
End;

Procedure TGLPolygon.Assign(Source: TPersistent);
Begin
   If Source Is TGLPolygon Then
   Begin
      FParts := TGLPolygon(Source).FParts;
   End;
   Inherited Assign(Source);
End;

Procedure TGLPolygon.BuildList(Var rci: TGLRenderContextInfo);
Var
   Normal: TAffineVector;
   pNorm: PAffineVector;
Begin
   If (Nodes.Count > 1) Then
   Begin
      Normal := Nodes.Normal;
      If VectorIsNull(Normal) Then
         pNorm := nil
      Else
         pNorm := @Normal;
      If ppTop In FParts Then
      Begin
         If SplineMode = lsmLines Then
            Nodes.RenderTesselatedPolygon(True, pNorm, 1)
         Else
            Nodes.RenderTesselatedPolygon(True, pNorm, Division);
      End;
      // tessellate bottom polygon
      If ppBottom In FParts Then
      Begin
         If Assigned(pNorm) Then
            NegateVector(Normal);
         If SplineMode = lsmLines Then
            Nodes.RenderTesselatedPolygon(True, pNorm, 1, True)
         Else
            Nodes.RenderTesselatedPolygon(True, pNorm, Division, True);
      End;
   End;
End;


Initialization


   RegisterClasses([TGLCylinder, TGLCone, TGLTorus, TGLDisk, TGLArrowLine, TGLAnnulus, TGLFrustrum, TGLPolygon, TGLCapsule, TGLArrowArc]);
End.
