{*************************************************************************
                PilotLogic Software House

  Package pl_GLScene
  This file is part of CodeTyphon Studio (https://www.pilotlogic.com/)

 ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 2.0
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GLScene Project
 *
 * ***** END LICENSE BLOCK *****
 ************************************************************************}

{
  Bitmap Fonts management classes
}
Unit GLBitmapFont;

{$I GLScene.inc}

Interface

Uses
   Classes,
   SysUtils,
   Types,
   Graphics,
   OpenGLTokens,
   GLScene,
   GLVectorGeometry,
   GLContext,
   GLCrossPlatform,
   GLTexture,
   GLState,
   GLUtils,
   GLGraphics,
   GLColor,
   GLBaseClasses,
   GLRenderContextInfo,
   GLTextureFormat,
   XOpenGL,
   GLVectorTypes,
   GLPersistentClasses;

Type

  {  An individual character range in a bitmap font.
    A range allows mapping ASCII characters to character tiles in a font
    bitmap, tiles are enumerated line then column (raster). }
   TGLBitmapFontRange = Class(TCollectionItem)
   Private
      Function GetStartASCII: WideString;
      Function GetStopASCII: WideString;
   Protected
      FStartASCII, FStopASCII: Widechar;
      FStartGlyphIdx, FStopGlyphIdx, FCharCount: Integer;
      Procedure SetStartASCII(Const val: WideString);
      Procedure SetStopASCII(Const val: WideString);
      Procedure SetStartGlyphIdx(val: Integer);
      Function GetDisplayName: String; Override;
   Public
      Constructor Create(Collection: TCollection); Override;
      Destructor Destroy; Override;
      Procedure Assign(Source: TPersistent); Override;
      Procedure NotifyChange;
   Published
      Property StartASCII: WideString read GetStartASCII write SetStartASCII;
      Property StopASCII: WideString read GetStopASCII write SetStopASCII;
      Property StartGlyphIdx: Integer read FStartGlyphIdx write SetStartGlyphIdx;
      Property StopGlyphIdx: Integer read FStopGlyphIdx;
      Property CharCount: Integer read FCharCount;
   End;

   TGLBitmapFontRanges = Class(TCollection)
   Private
      FCharCount: Integer;
   Protected
      FOwner: TComponent;
      Function GetOwner: TPersistent; Override;
      Procedure SetItems(index: Integer; Const val: TGLBitmapFontRange);
      Function GetItems(index: Integer): TGLBitmapFontRange;
      Function CalcCharacterCount: Integer;
      Procedure Update(Item: TCollectionItem); Override;
   Public
      Constructor Create(AOwner: TComponent);
      Destructor Destroy; Override;
      Function Add: TGLBitmapFontRange; Overload;
      Function Add(Const StartASCII, StopASCII: Widechar): TGLBitmapFontRange; Overload;
      Function Add(Const StartASCII, StopASCII: Ansichar): TGLBitmapFontRange; Overload;
      Function FindItemID(ID: Integer): TGLBitmapFontRange;
      Property Items[index: Integer]: TGLBitmapFontRange read GetItems write SetItems; Default;
    {  Converts an ASCII character into a tile index.
      Return -1 if character cannot be rendered. }
      Function CharacterToTileIndex(aChar: Widechar): Integer;
      Function TileIndexToChar(aIndex: Integer): Widechar;
      Procedure NotifyChange;
      // Total number of characters in the ranges; cached for performance
      Property CharacterCount: Integer read FCharCount;
   End;

   PCharInfo = ^TCharInfo;

   TCharInfo = Record
      l, t, w: Word;
   End;

  {  Provides access to individual characters in a BitmapFont.
    Only fixed-width bitmap fonts are supported, the characters are enumerated
    in a raster fashion (line then column).
    Transparency is all or nothing, the transparent color being that of the
    top left pixel of the Glyphs bitmap.
    Performance note: as usual, for best performance, you base font bitmap
    dimensions should be close to a power of two, and have at least 1 pixel
    spacing between characters (horizontally and vertically) to avoid artefacts
    when rendering with linear filtering. }
   TGLCustomBitmapFont = Class(TGLUpdateAbleComponent)
   Private
      FRanges: TGLBitmapFontRanges;
      FGlyphs: TPicture;
      FCharWidth, FCharHeight: Integer;
      FGlyphsIntervalX, FGlyphsIntervalY: Integer;
      FHSpace, FVSpace, FHSpaceFix: Integer;
      FUsers: TList;
      FMinFilter: TGLMinFilter;
      FMagFilter: TGLMagFilter;
      FTextureWidth, FTextureHeight: Integer;
      FTextRows, FTextCols: Integer;
      FGlyphsAlpha: TGLTextureImageAlpha;
      FTextures: TList;
      FTextureModified: Boolean;
      FLastTexture: TGLTextureHandle;
   Protected
      FChars: Array Of TCharInfo;
      FCharsLoaded: Boolean;
      //=== ct9999 ===========
      FCharRects: Array Of TVector;
      Procedure SetCharRects(index: Integer; Const p: TVector);
      //======================
      Procedure ResetCharWidths(w: Integer = -1);
      Procedure SetCharWidths(index, Value: Integer);
      Procedure SetRanges(Const val: TGLBitmapFontRanges);
      Procedure SetGlyphs(Const val: TPicture);
      Procedure SetCharWidth(Const val: Integer);
      Procedure SetCharHeight(Const val: Integer);
      Procedure SetGlyphsIntervalX(Const val: Integer);
      Procedure SetGlyphsIntervalY(Const val: Integer);
      Procedure OnGlyphsChanged(Sender: TObject);
      Procedure SetHSpace(Const val: Integer);
      Procedure SetVSpace(Const val: Integer);
      Procedure SetMagFilter(AValue: TGLMagFilter);
      Procedure SetMinFilter(AValue: TGLMinFilter);
      Procedure SetGlyphsAlpha(val: TGLTextureImageAlpha);
      Procedure TextureChanged;
      Procedure FreeTextureHandle; Virtual;
      Function TextureFormat: Integer; Virtual;
      Procedure InvalidateUsers;
      Function CharactersPerRow: Integer;
      Procedure GetCharTexCoords(Ch: Widechar; Var TopLeft, BottomRight: TTexPoint);
      Procedure GetICharTexCoords(Var ARci: TGLRenderContextInfo; Chi: Integer; out TopLeft, BottomRight: TTexPoint);
      Procedure PrepareImage(Var ARci: TGLRenderContextInfo); Virtual;
      Procedure PrepareParams(Var ARci: TGLRenderContextInfo);
    {  A single bitmap containing all the characters.
      The transparent color is that of the top left pixel. }
      Property Glyphs: TPicture read FGlyphs write SetGlyphs;
      {  Nb of horizontal pixels between two columns in the Glyphs. }
      Property GlyphsIntervalX: Integer read FGlyphsIntervalX write SetGlyphsIntervalX;
      {  Nb of vertical pixels between two rows in the Glyphs. }
      Property GlyphsIntervalY: Integer read FGlyphsIntervalY write SetGlyphsIntervalY;
    {  Ranges allow converting between ASCII and tile indexes.
      See TGLCustomBitmapFontRange. }
      Property Ranges: TGLBitmapFontRanges read FRanges write SetRanges;
      {  Width of a single character. }
      Property CharWidth: Integer read FCharWidth write SetCharWidth Default 16;
      {  Pixels in between rendered characters (horizontally). }
      Property HSpace: Integer read FHSpace write SetHSpace Default 1;
      {  Pixels in between rendered lines (vertically). }
      Property VSpace: Integer read FVSpace write SetVSpace Default 1;
    {  Horizontal spacing fix offset.
      This property is for internal use, and is added to the hspacing
      of each character when rendering, typically to fix extra spacing. }
      Property HSpaceFix: Integer read FHSpaceFix write FHSpaceFix;
      Property MagFilter: TGLMagFilter read FMagFilter write SetMagFilter Default maLinear;
      Property MinFilter: TGLMinFilter read FMinFilter write SetMinFilter Default miLinear;
      Property GlyphsAlpha: TGLTextureImageAlpha read FGlyphsAlpha write FGlyphsAlpha Default tiaDefault;
   Public
      Constructor Create(AOwner: TComponent); Override;
      Destructor Destroy; Override;
      Procedure RegisterUser(anObject: TGLBaseSceneObject); Virtual;
      Procedure UnRegisterUser(anObject: TGLBaseSceneObject); Virtual;
    {  Renders the given string at current position or at position given by the optional position variable.
      The current matrix is blindly used, meaning you can render all kinds
      of rotated and linear distorted text with this method, OpenGL
      Enable states are also possibly altered. }
      Procedure RenderString(Var ARci: TGLRenderContextInfo; Const aText: Unicodestring; aAlignment: TAlignment; aLayout: TGLTextLayout; Const aColor: TColorVector; aPosition: PVector = nil;
         aReverseY: Boolean = False); Overload; Virtual;
    {  A simpler canvas-style TextOut helper for RenderString.
      The rendering is reversed along Y by default, to allow direct use
      with TGLCanvas }
      Procedure TextOut(Var rci: TGLRenderContextInfo; X, Y: Single; Const Text: Unicodestring; Const Color: TColorVector); Overload;
      Procedure TextOut(Var rci: TGLRenderContextInfo; X, Y: Single; Const Text: Unicodestring; Const Color: TColor); Overload;
      Function TextWidth(Const Text: Unicodestring): Integer;
      Function CharacterToTileIndex(aChar: Widechar): Integer; Virtual;
      Function TileIndexToChar(aIndex: Integer): Widechar; Virtual;
      Function CharacterCount: Integer; Virtual;
      {  Get the actual width for this char. }
      Function GetCharWidth(Ch: Widechar): Integer;
      {  Get the actual pixel width for this string. }
      Function CalcStringWidth(Const aText: Unicodestring): Integer; Overload; Virtual;
      // make texture if needed
      Procedure CheckTexture(Var ARci: TGLRenderContextInfo);
      {  Height of a single character. }
      Property CharHeight: Integer read FCharHeight write SetCharHeight Default 16;
      Property TextureWidth: Integer read FTextureWidth write FTextureWidth;
      Property TextureHeight: Integer read FTextureHeight write FTextureHeight;
   End;

  {  See TGLCustomBitmapFont.
    This class only publuishes some of the properties. }
   TGLBitmapFont = Class(TGLCustomBitmapFont)
   Published
      Property Glyphs;
      Property GlyphsIntervalX;
      Property GlyphsIntervalY;
      Property Ranges;
      Property CharWidth;
      Property CharHeight;
      Property HSpace;
      Property VSpace;
      Property MagFilter;
      Property MinFilter;
      Property GlyphsAlpha;
   End;

   TGLFlatTextOption = (ftoTwoSided);
   TGLFlatTextOptions = Set Of TGLFlatTextOption;

  {  A 2D text displayed and positionned in 3D coordinates.
    The FlatText uses a character font defined and stored by a TGLBitmapFont
    component. Default character scale is 1 font pixel = 1 space unit. }
   TGLFlatText = Class(TGLImmaterialSceneObject)
   Private
      FBitmapFont: TGLCustomBitmapFont;
      FText: Unicodestring;
      FAlignment: TAlignment;
      FLayout: TGLTextLayout;
      FModulateColor: TGLColor;
      FOptions: TGLFlatTextOptions;
   Protected
      Procedure SetBitmapFont(Const val: TGLCustomBitmapFont);
      Procedure SetText(Const val: Unicodestring);
      Procedure SetAlignment(Const val: TAlignment);
      Procedure SetLayout(Const val: TGLTextLayout);
      Procedure SetModulateColor(Const val: TGLColor);
      Procedure SetOptions(Const val: TGLFlatTextOptions);
      Procedure Notification(AComponent: TComponent; Operation: TOperation); Override;
   Public
      Constructor Create(AOwner: TComponent); Override;
      Destructor Destroy; Override;
      Procedure DoRender(Var rci: TGLRenderContextInfo; aRenderSelf, aRenderChildren: Boolean); Override;
      Procedure Assign(Source: TPersistent); Override;
   Published
    {  Refers the bitmap font to use.
      The referred bitmap font component stores and allows access to
      individual character bitmaps. }
      Property BitmapFont: TGLCustomBitmapFont read FBitmapFont write SetBitmapFont;
    {  Text to render.
      Be aware that only the characters available in the bitmap font will
      be rendered. CR LF sequences are allowed. }
      Property Text: Unicodestring read FText write SetText;
    {  Controls the text alignment (horizontal).
      Possible values : taLeftJustify, taRightJustify, taCenter }
      Property Alignment: TAlignment read FAlignment write SetAlignment;
    {  Controls the text layout (vertical).
      Possible values : tlTop, tlCenter, tlBottom }
      Property Layout: TGLTextLayout read FLayout write SetLayout;
      {  Color modulation, can be used for fade in/out too. }
      Property ModulateColor: TGLColor read FModulateColor write SetModulateColor;
    {  Flat text options.
       ftoTwoSided : when set the text will be visible from its two
      sides even if faceculling is on (at the scene-level).
       }
      Property Options: TGLFlatTextOptions read FOptions write SetOptions;
   End;


Implementation



// ------------------
// ------------------ TGLBitmapFontRange ------------------
// ------------------

Constructor TGLBitmapFontRange.Create(Collection: TCollection);
Begin
   Inherited Create(Collection);
End;

Destructor TGLBitmapFontRange.Destroy;
Begin
   Inherited;
End;

Procedure TGLBitmapFontRange.Assign(Source: TPersistent);
Begin
   If Source Is TGLBitmapFontRange Then
   Begin
      FStartASCII := TGLBitmapFontRange(Source).FStartASCII;
      FStopASCII := TGLBitmapFontRange(Source).FStopASCII;
      FStartGlyphIdx := TGLBitmapFontRange(Source).FStartGlyphIdx;
      NotifyChange;
   End Else
      Inherited;
End;

Procedure TGLBitmapFontRange.NotifyChange;
Begin
   FCharCount := Integer(FStopASCII) - Integer(FStartASCII) + 1;
   FStopGlyphIdx := FStartGlyphIdx + FCharCount - 1;
   If Assigned(Collection) Then
      (Collection As TGLBitmapFontRanges).NotifyChange;
End;

Function TGLBitmapFontRange.GetDisplayName: String;
Begin
   Result := Format('ASCII [#%d, #%d] -> Glyphs [%d, %d]', [Integer(FStartASCII), Integer(FStopASCII), StartGlyphIdx, StopGlyphIdx]);
End;

Function TGLBitmapFontRange.GetStartASCII: WideString;
Begin
   Result := FStartASCII;
End;

Function TGLBitmapFontRange.GetStopASCII: WideString;
Begin
   Result := FStopASCII;
End;

Procedure TGLBitmapFontRange.SetStartASCII(Const val: WideString);
Begin
   If (Length(val) > 0) And (val[1] <> FStartASCII) Then
   Begin
      FStartASCII := val[1];
      If FStartASCII > FStopASCII Then
         FStopASCII := FStartASCII;
      NotifyChange;
   End;
End;

Procedure TGLBitmapFontRange.SetStopASCII(Const val: WideString);
Begin
   If (Length(val) > 0) And (FStopASCII <> val[1]) Then
   Begin
      FStopASCII := val[1];
      If FStopASCII < FStartASCII Then
         FStartASCII := FStopASCII;
      NotifyChange;
   End;
End;

Procedure TGLBitmapFontRange.SetStartGlyphIdx(val: Integer);
Begin
   val := MaxInteger(0, val);
   If val <> FStartGlyphIdx Then
   Begin
      FStartGlyphIdx := val;
      NotifyChange;
   End;
End;

// ------------------
// ------------------ TGLBitmapFontRanges ------------------
// ------------------
Constructor TGLBitmapFontRanges.Create(AOwner: TComponent);
Begin
   FOwner := AOwner;
   Inherited Create(TGLBitmapFontRange);
End;

Destructor TGLBitmapFontRanges.Destroy;
Begin
   Inherited;
End;

Function TGLBitmapFontRanges.GetOwner: TPersistent;
Begin
   Result := FOwner;
End;

Procedure TGLBitmapFontRanges.SetItems(index: Integer; Const val: TGLBitmapFontRange);
Begin
   Inherited Items[index] := val;
End;

Function TGLBitmapFontRanges.GetItems(index: Integer): TGLBitmapFontRange;
Begin
   Result := TGLBitmapFontRange(Inherited Items[index]);
End;

Function TGLBitmapFontRanges.Add: TGLBitmapFontRange;
Begin
   Result := (Inherited Add) As TGLBitmapFontRange;
End;

Function TGLBitmapFontRanges.Add(Const StartASCII, StopASCII: Widechar): TGLBitmapFontRange;
Begin
   Result := Add;
   Result.StartASCII := StartASCII;
   Result.StopASCII := StopASCII;
End;

Function TGLBitmapFontRanges.Add(Const StartASCII, StopASCII: Ansichar): TGLBitmapFontRange;
Begin
   Result := Add(CharToWideChar(StartASCII), CharToWideChar(StopASCII));
End;

Function TGLBitmapFontRanges.FindItemID(ID: Integer): TGLBitmapFontRange;
Begin
   Result := (Inherited FindItemID(ID)) As TGLBitmapFontRange;
End;

Function TGLBitmapFontRanges.CharacterToTileIndex(aChar: Widechar): Integer;
Var
   i: Integer;
Begin
   Result := -1;
   For i := 0 To Count - 1 Do
      With Items[i] Do
      Begin
         If (aChar >= FStartASCII) And (aChar <= FStopASCII) Then
         Begin
            Result := StartGlyphIdx + Integer(aChar) - Integer(FStartASCII);
            Break;
         End;
      End;
End;

Function TGLBitmapFontRanges.TileIndexToChar(aIndex: Integer): Widechar;
Var
   i: Integer;
Begin
   Result := #0;
   For i := 0 To Count - 1 Do
      With Items[i] Do
      Begin
         If (aIndex >= StartGlyphIdx) And (aIndex <= StopGlyphIdx) Then
         Begin
            Result := Widechar(aIndex - StartGlyphIdx + Integer(FStartASCII));
            Break;
         End;
      End;
End;

Procedure TGLBitmapFontRanges.Update(Item: TCollectionItem);
Begin
   Inherited;
   NotifyChange;
End;

Procedure TGLBitmapFontRanges.NotifyChange;
Begin
   FCharCount := CalcCharacterCount;

   If Assigned(FOwner) Then
   Begin
      If FOwner Is TGLBaseSceneObject Then
         TGLBaseSceneObject(FOwner).StructureChanged
      Else
      If FOwner Is TGLCustomBitmapFont Then
         TGLCustomBitmapFont(FOwner).NotifyChange(Self);
   End;
End;

Function TGLBitmapFontRanges.CalcCharacterCount: Integer;
Var
   i: Integer;
Begin
   Result := 0;
   For i := 0 To Count - 1 Do
      With Items[i] Do
         Inc(Result, Integer(FStopASCII) - Integer(FStartASCII) + 1);
End;

// ------------------
// ------------------ TGLCustomBitmapFont ------------------
// ------------------
Constructor TGLCustomBitmapFont.Create(AOwner: TComponent);
Begin
   Inherited Create(AOwner);
   FRanges := TGLBitmapFontRanges.Create(Self);
   FGlyphs := TPicture.Create;
   FGlyphs.OnChange := OnGlyphsChanged;
   FCharWidth := 16;
   FCharHeight := 16;
   FHSpace := 1;
   FVSpace := 1;
   FUsers := TList.Create;
   FMinFilter := miLinear;
   FMagFilter := maLinear;
   FTextures := TList.Create;
   FTextureModified := True;
End;

Destructor TGLCustomBitmapFont.Destroy;
Begin
   FreeTextureHandle;
   Inherited Destroy;
   FTextures.Free;
   FRanges.Free;
   FGlyphs.Free;
   Assert(FUsers.Count = 0);
   FUsers.Free;
End;

Function TGLCustomBitmapFont.GetCharWidth(Ch: Widechar): Integer;
Var
   chi: Integer;
Begin
   chi := CharacterToTileIndex(ch);
   If Length(FChars) = 0 Then
      ResetCharWidths;
   If chi >= 0 Then
      Result := FChars[chi].w
   Else
      Result := 0;
End;

Function TGLCustomBitmapFont.CalcStringWidth(Const aText: Unicodestring): Integer;
Var
   i: Integer;
Begin
   If aText <> '' Then
   Begin
      Result := -HSpace + Length(aText) * (HSpaceFix + HSpace);
      For i := 1 To Length(aText) Do
         Result := Result + GetCharWidth(aText[i]);
   End Else
      Result := 0;
End;

Procedure TGLCustomBitmapFont.ResetCharWidths(w: Integer = -1);
Var
   i: Integer;
Begin
   FCharsLoaded := False;
   i := CharacterCount;
   If Length(FChars) < i Then
      SetLength(FChars, i);
   If w < 0 Then
      w := CharWidth;
   For i := 0 To High(FChars) Do
      FChars[i].w := w;
End;

Procedure TGLCustomBitmapFont.SetCharWidths(index, Value: Integer);
Begin
   If index >= 0 Then
      FChars[index].w := Value;
End;

Procedure TGLCustomBitmapFont.SetRanges(Const val: TGLBitmapFontRanges);
Begin
   FRanges.Assign(val);
   InvalidateUsers;
End;

Procedure TGLCustomBitmapFont.SetGlyphs(Const val: TPicture);
Begin
   FGlyphs.Assign(val);
End;

Procedure TGLCustomBitmapFont.SetCharWidth(Const val: Integer);
Begin
   If val <> FCharWidth Then
   Begin
      If val > 1 Then
         FCharWidth := val
      Else
         FCharWidth := 1;
      InvalidateUsers;
   End;
End;

Procedure TGLCustomBitmapFont.SetCharHeight(Const val: Integer);
Begin
   If val <> FCharHeight Then
   Begin
      If val > 1 Then
         FCharHeight := val
      Else
         FCharHeight := 1;
      InvalidateUsers;
   End;
End;

Procedure TGLCustomBitmapFont.SetGlyphsIntervalX(Const val: Integer);
Begin
   If val > 0 Then
      FGlyphsIntervalX := val
   Else
      FGlyphsIntervalX := 0;
   InvalidateUsers;
End;

Procedure TGLCustomBitmapFont.SetGlyphsIntervalY(Const val: Integer);
Begin
   If val > 0 Then
      FGlyphsIntervalY := val
   Else
      FGlyphsIntervalY := 0;
   InvalidateUsers;
End;

Procedure TGLCustomBitmapFont.SetHSpace(Const val: Integer);
Begin
   If val <> FHSpace Then
   Begin
      FHSpace := val;
      InvalidateUsers;
   End;
End;

Procedure TGLCustomBitmapFont.SetVSpace(Const val: Integer);
Begin
   If val <> FVSpace Then
   Begin
      FVSpace := val;
      InvalidateUsers;
   End;
End;

Procedure TGLCustomBitmapFont.SetMagFilter(AValue: TGLMagFilter);
Begin
   If AValue <> FMagFilter Then
   Begin
      FMagFilter := AValue;
      TextureChanged;
      InvalidateUsers;
   End;
End;

Procedure TGLCustomBitmapFont.SetMinFilter(AValue: TGLMinFilter);
Begin
   If AValue <> FMinFilter Then
   Begin
      FMinFilter := AValue;
      TextureChanged;
      InvalidateUsers;
   End;
End;

Procedure TGLCustomBitmapFont.SetGlyphsAlpha(val: TGLTextureImageAlpha);
Begin
   If val <> FGlyphsAlpha Then
   Begin
      FGlyphsAlpha := val;
      TextureChanged;
      InvalidateUsers;
   End;
End;

Procedure TGLCustomBitmapFont.OnGlyphsChanged(Sender: TObject);
Begin
   InvalidateUsers;
   // when empty, width is 0 and roundup give 1
   If Not Glyphs.Graphic.Empty Then
   Begin
      If FTextureWidth = 0 Then
         FTextureWidth := RoundUpToPowerOf2(Glyphs.Width);
      If FTextureHeight = 0 Then
         FTextureHeight := RoundUpToPowerOf2(Glyphs.Height);
   End;
End;

Procedure TGLCustomBitmapFont.RegisterUser(anObject: TGLBaseSceneObject);
Begin
   Assert(FUsers.IndexOf(anObject) < 0);
   FUsers.Add(anObject);
End;

Procedure TGLCustomBitmapFont.UnRegisterUser(anObject: TGLBaseSceneObject);
Begin
   FUsers.Remove(anObject);
End;

Procedure TGLCustomBitmapFont.PrepareImage(Var ARci: TGLRenderContextInfo);
Var
   Stream: TMemoryStream;
   Glyphs2: TPicture ;
   bitmap: TBitmap;
   bitmap32: TGLImage;
   cap: Integer;
   X, Y, w, h: Integer;
   t: TGLTextureHandle;
Begin
   // only check when really used
   If FTextureWidth = 0 Then
   Begin
      FTextureWidth := ARci.GLStates.MaxTextureSize;
      If FTextureWidth > 512 Then
         FTextureWidth := 512;
      If FTextureWidth < 64 Then
         FTextureWidth := 64;
   End;
   If FTextureHeight = 0 Then
   Begin
      FTextureHeight := ARci.GLStates.MaxTextureSize;
      If FTextureHeight > 512 Then
         FTextureHeight := 512;
      If FTextureHeight < 64 Then
         FTextureHeight := 64;
   End;

   X := 0;
   Y := 0;
   w := Glyphs.Width;
   h := Glyphs.Height;

   // was an error...
   FTextRows := 1 + (h - 1) Div FTextureHeight;
   FTextCols := 1 + (w - 1) Div FTextureWidth;

   bitmap := TBitmap.Create;
   With bitmap Do
   Begin
      //=== ct9999 ========
      PixelFormat := glpf32bit;
      //===================
      SetSize(RoundUpToPowerOf2(FTextureWidth),
         RoundUpToPowerOf2(FTextureHeight));
   End;
   bitmap32 := TGLImage.Create;
   While (X < w) And (Y < h) Do
   Begin
      t := TGLTextureHandle.Create;
      FTextures.Add(t);
      // prepare handle
      t.AllocateHandle;
      // texture registration
      t.Target := ttTexture2D;
      ARci.GLStates.TextureBinding[0, ttTexture2D] := t.Handle;

      Glyphs2:=TPicture.Create;
      Stream:= TMemoryStream.Create;
      Glyphs.SaveToStreamWithFileExt(Stream,'bmp');
      Stream.Position:=0;;
      Glyphs2.LoadFromStreamWithFileExt(Stream,'bmp');
      // copy data
      bitmap.Canvas.Draw(-X, -Y, Glyphs2.Graphic);
      FreeAndNil(Glyphs2);
      FreeAndNil(Stream);

      // Clipboard.Assign(bitmap);
      bitmap32.Assign(bitmap);
      bitmap32.Narrow;
      With bitmap32 Do
      Begin
         Case FGlyphsAlpha Of
            tiaAlphaFromIntensity:
               SetAlphaFromIntensity;
            tiaSuperBlackTransparent:
               SetAlphaTransparentForColor($000000);
            tiaLuminance:
               SetAlphaFromIntensity;
            tiaLuminanceSqrt:
            Begin
               SetAlphaFromIntensity;
               SqrtAlpha;
            End;
            tiaOpaque:
               SetAlphaToValue(255);
            tiaDefault, tiaTopLeftPointColorTransparent:
               SetAlphaTransparentForColor(Data[Width * (Height - 1)]);
            Else Assert(False);
         End;
         RegisterAsOpenGLTexture(t, Not (FMinFilter In [miNearest, miLinear]),
            TextureFormat, cap, cap, cap);
      End;

      PrepareParams(ARci);
      t.NotifyDataUpdated;

      Inc(X, FTextureWidth);
      If X >= w Then
      Begin
         Inc(Y, FTextureHeight);
         X := 0;
      End;
   End;
   bitmap.Free;
   bitmap32.Free;
End;

Procedure TGLCustomBitmapFont.PrepareParams(Var ARci: TGLRenderContextInfo);
Const
   cTextureMagFilter: Array [maNearest .. maLinear] Of Cardinal = (GL_NEAREST, GL_LINEAR);
   cTextureMinFilter: Array [miNearest .. miLinearMipmapLinear] Of Cardinal =
      (GL_NEAREST, GL_LINEAR, GL_NEAREST_MIPMAP_NEAREST, GL_LINEAR_MIPMAP_NEAREST,
      GL_NEAREST_MIPMAP_LINEAR, GL_LINEAR_MIPMAP_LINEAR);
Begin
   With ARci.GLStates Do
   Begin
      UnpackAlignment := 4;
      UnpackRowLength := 0;
      UnpackSkipRows := 0;
      UnpackSkipPixels := 0;
   End;

   With GL Do
   Begin
      Hint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

      TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
      TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

      TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, cTextureMinFilter[FMinFilter]);
      TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, cTextureMagFilter[FMagFilter]);
   End;
End;

Function TGLCustomBitmapFont.TileIndexToChar(aIndex: Integer): Widechar;
Begin
   Result := FRanges.TileIndexToChar(aIndex);
End;

Function TGLCustomBitmapFont.CharacterToTileIndex(aChar: Widechar): Integer;
Begin
   Result := FRanges.CharacterToTileIndex(aChar);
End;

Procedure TGLCustomBitmapFont.RenderString(Var ARci: TGLRenderContextInfo; Const aText: Unicodestring; aAlignment: TAlignment; aLayout: TGLTextLayout; Const aColor: TColorVector; aPosition: PVector = nil; aReverseY: Boolean = False);

   Function AlignmentAdjustement(p: Integer): Single;
   Var
      i: Integer;
   Begin
      i := 0;
      While (p <= Length(aText)) And (aText[p] <> #13) Do
      Begin
         Inc(p);
         Inc(i);
      End;
      Case aAlignment Of
         taLeftJustify:
            Result := 0;
         taRightJustify:
            Result := -CalcStringWidth(Copy(aText, p - i, i))
         Else // taCenter
            Result := Round(-CalcStringWidth(Copy(aText, p - i, i)) * 0.5);
      End;
   End;

   Function LayoutAdjustement: Single;
   Var
      i, n: Integer;
   Begin
      n := 1;
      For i := 1 To Length(aText) Do
         If aText[i] = #13 Then
            Inc(n);
      Case TGLTextLayout(aLayout) Of
         tlTop: Result := 0;
         tlBottom: Result := (n * (CharHeight + VSpace) - VSpace);
         Else // tlCenter
            Result := Round((n * (CharHeight + VSpace) - VSpace) * 0.5);
      End;
   End;

Var
   i, chi: Integer;
   pch: PCharInfo;
   TopLeft, BottomRight: TTexPoint;
   vTopLeft, vBottomRight: TVector;
   deltaV, spaceDeltaH: Single;
   currentChar: Widechar;
Begin
   If (aText = '') Then
      Exit;
   // prepare texture if necessary
   CheckTexture(ARci);
   // precalcs
   If Assigned(aPosition) Then
      MakePoint(vTopLeft, aPosition.X + AlignmentAdjustement(1),
         aPosition.Y + LayoutAdjustement, 0)
   Else
      MakePoint(vTopLeft, AlignmentAdjustement(1), LayoutAdjustement, 0);
   deltaV := -(CharHeight + VSpace);
   If aReverseY Then
      vBottomRight.Y := vTopLeft.Y + CharHeight
   Else
      vBottomRight.Y := vTopLeft.Y - CharHeight;
   vBottomRight.Z := 0;
   vBottomRight.W := 1;
   spaceDeltaH := GetCharWidth(#32) + HSpaceFix + HSpace;
   // set states
   With ARci.GLStates Do
   Begin
      ActiveTextureEnabled[ttTexture2D] := True;
      Disable(stLighting);
      Enable(stBlend);
      SetBlendFunc(bfSrcAlpha, bfOneMinusSrcAlpha);
      FLastTexture := nil;
   End;

   // start rendering
   gl.Color4fv(@aColor);
   gl.Begin_(GL_QUADS);
   For i := 1 To Length(aText) Do
   Begin
      currentChar := Widechar(aText[i]);
      Case currentChar Of
         #0 .. #12, #14 .. #31:
            ; // ignore
         #13:
         Begin
            If Assigned(aPosition) Then
               vTopLeft.X := aPosition.X + AlignmentAdjustement(i + 1)
            Else
               vTopLeft.X := AlignmentAdjustement(i + 1);
            vTopLeft.Y := vTopLeft.Y + deltaV;
            If aReverseY Then
               vBottomRight.Y := vTopLeft.Y + CharHeight
            Else
               vBottomRight.Y := vTopLeft.Y - CharHeight;
         End;
         #32:
            vTopLeft.X := vTopLeft.X + spaceDeltaH;
         Else chi := CharacterToTileIndex(currentChar);
            If chi < 0 Then
               continue; // not found
            pch := @FChars[chi];
            If pch.w > 0 Then
            Begin
               GetICharTexCoords(ARci, chi, TopLeft, BottomRight);
               vBottomRight.X := vTopLeft.X + pch.w;

               gl.TexCoord2fv(@TopLeft);
               gl.Vertex4fv(@vTopLeft);

               gl.TexCoord2f(TopLeft.S, BottomRight.t);
               gl.Vertex2f(vTopLeft.X, vBottomRight.Y);

               gl.TexCoord2fv(@BottomRight);
               gl.Vertex4fv(@vBottomRight);

               gl.TexCoord2f(BottomRight.S, TopLeft.t);
               gl.Vertex2f(vBottomRight.X, vTopLeft.Y);

               vTopLeft.X := vTopLeft.X + pch.w + HSpace;
            End;
      End;
   End;
   gl.End_;
   // unbind texture
   ARci.GLStates.TextureBinding[0, ttTexture2D] := 0;
   ARci.GLStates.ActiveTextureEnabled[ttTexture2D] := False;
End;

Procedure TGLCustomBitmapFont.TextOut(Var rci: TGLRenderContextInfo; X, Y: Single; Const Text: Unicodestring; Const Color: TColorVector);
Var
   V: TVector;
Begin
   V.X := X;
   V.Y := Y;
   V.Z := 0;
   V.W := 1;
   RenderString(rci, Text, taLeftJustify, tlTop, Color, @V, True);
End;

Procedure TGLCustomBitmapFont.TextOut(Var rci: TGLRenderContextInfo; X, Y: Single; Const Text: Unicodestring; Const Color: TColor);
Begin
   TextOut(rci, X, Y, Text, ConvertWinColor(Color));
End;

Function TGLCustomBitmapFont.TextWidth(Const Text: Unicodestring): Integer;
Begin
   Result := CalcStringWidth(Text);
End;

Function TGLCustomBitmapFont.CharactersPerRow: Integer;
Begin
   If FGlyphs.Width > 0 Then
      Result := (FGlyphs.Width + FGlyphsIntervalX) Div (FGlyphsIntervalX + FCharWidth)
   Else
      Result := 0;
End;

Function TGLCustomBitmapFont.CharacterCount: Integer;
Begin
   Result := FRanges.CharacterCount;
End;

Procedure TGLCustomBitmapFont.GetCharTexCoords(Ch: Widechar; Var TopLeft, BottomRight: TTexPoint);
Var
   chi, tileIndex: Integer;
   ci: TCharInfo;
   r: Integer;
Begin
   chi := CharacterToTileIndex(ch);
   If Not FCharsLoaded Then
   Begin
      ResetCharWidths;
      FCharsLoaded := True;
      r := CharactersPerRow;
      For tileIndex := 0 To CharacterCount - 1 Do
      Begin
         FChars[tileIndex].l := (tileIndex Mod r) * (CharWidth + GlyphsIntervalX);
         FChars[tileIndex].t := (tileIndex Div r) * (CharHeight + GlyphsIntervalY);
      End;
   End;

   If (chi < 0) Or (chi >= CharacterCount) Then
   Begin
      // invalid char
      TopLeft := NullTexPoint;
      BottomRight := NullTexPoint;
      Exit;
   End;

   ci := FChars[chi];
   ci.l := ci.l Mod FTextureWidth;
   ci.t := ci.t Mod FTextureHeight;

   TopLeft.S := ci.l / FTextureWidth;
   TopLeft.t := 1 - ci.t / FTextureHeight;
   BottomRight.S := (ci.l + ci.w) / FTextureWidth;
   BottomRight.t := 1 - (ci.t + CharHeight) / FTextureHeight;
End;

// TileIndexToTexCoords it also activates the target texture
Procedure TGLCustomBitmapFont.GetICharTexCoords(Var ARci: TGLRenderContextInfo; Chi: Integer; out TopLeft, BottomRight: TTexPoint);
Var
   tileIndex: Integer;
   ci: TCharInfo;
   t: TGLTextureHandle;
   r, c: Integer;
Begin
   If Not FCharsLoaded Then
   Begin
      r := CharactersPerRow;
      If r = 0 Then
         Exit;
      ResetCharWidths;
      FCharsLoaded := True;
      For tileIndex := 0 To CharacterCount - 1 Do
      Begin
         FChars[tileIndex].l := (tileIndex Mod r) * (CharWidth + GlyphsIntervalX);
         FChars[tileIndex].t := (tileIndex Div r) * (CharHeight + GlyphsIntervalY);
      End;
   End;

   If (chi < 0) Or (chi >= CharacterCount) Then
   Begin
      // invalid char
      TopLeft := NullTexPoint;
      BottomRight := NullTexPoint;
      Exit;
   End;

   ci := FChars[chi];

   c := ci.l Div FTextureWidth;
   r := ci.t Div FTextureHeight;
   ci.l := ci.l Mod FTextureWidth;
   ci.t := ci.t Mod FTextureHeight;
   t := FTextures[r * FTextCols + c];

   TopLeft.S := ci.l / FTextureWidth;
   TopLeft.t := 1 - ci.t / FTextureHeight;
   BottomRight.S := (ci.l + ci.w) / FTextureWidth;
   BottomRight.t := 1 - (ci.t + CharHeight) / FTextureHeight;

   If t <> FLastTexture Then
   Begin
      FLastTexture := t;
      gl.End_;
      ARci.GLStates.TextureBinding[0, ttTexture2D] := t.Handle;
      gl.TexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
      gl.Begin_(GL_QUADS);
   End;
End;

Procedure TGLCustomBitmapFont.InvalidateUsers;
Var
   i: Integer;
Begin
   FCharsLoaded := False;
   FTextureModified := True;
   For i := FUsers.Count - 1 Downto 0 Do
      TGLBaseSceneObject(FUsers[i]).NotifyChange(Self);
End;

Procedure TGLCustomBitmapFont.FreeTextureHandle;
Var
   i: Integer;
Begin
   FTextureModified := True;
   For i := 0 To FTextures.Count - 1 Do
      TObject(FTextures[i]).Free;
   FTextures.Clear;
End;

Procedure TGLCustomBitmapFont.TextureChanged;
Begin
   FTextureModified := True;
End;

// force texture when needed
Procedure TGLCustomBitmapFont.CheckTexture(Var ARci: TGLRenderContextInfo);
Var
   i: Integer;
Begin
   // important: IsDataNeedUpdate might come from another source!
   For i := 0 To FTextures.Count - 1 Do
      FTextureModified := FTextureModified Or TGLTextureHandle(FTextures[i]).IsDataNeedUpdate;

   If FTextureModified Then
   Begin
      FreeTextureHandle; // instances are recreated in prepare
      PrepareImage(ARci);
      FTextureModified := False;
   End;
End;

Function TGLCustomBitmapFont.TextureFormat: Integer;
Begin
   Result := GL_RGBA;
End;

// ------------------
// ------------------ TGLFlatText ------------------
// ------------------
Constructor TGLFlatText.Create(AOwner: TComponent);
Begin
   Inherited;
   ObjectStyle := ObjectStyle + [osDirectDraw, osNoVisibilityCulling];
   FModulateColor := TGLColor.CreateInitialized(Self, clrWhite);
End;

Destructor TGLFlatText.Destroy;
Begin
   FModulateColor.Free;
   BitmapFont := nil;
   Inherited;
End;

Procedure TGLFlatText.Notification(AComponent: TComponent; Operation: TOperation);
Begin
   If (Operation = opRemove) And (AComponent = FBitmapFont) Then
      BitmapFont := nil;
   Inherited;
End;

Procedure TGLFlatText.SetBitmapFont(Const val: TGLCustomBitmapFont);
Begin
   If val <> FBitmapFont Then
   Begin
      If Assigned(FBitmapFont) Then
         FBitmapFont.UnRegisterUser(Self);
      FBitmapFont := val;
      If Assigned(FBitmapFont) Then
      Begin
         FBitmapFont.RegisterUser(Self);
         FBitmapFont.FreeNotification(Self);
      End;
      StructureChanged;
   End;
End;

Procedure TGLFlatText.SetText(Const val: Unicodestring);
Begin
   FText := val;
   StructureChanged;
End;

Procedure TGLFlatText.SetAlignment(Const val: TAlignment);
Begin
   FAlignment := val;
   StructureChanged;
End;

Procedure TGLFlatText.SetLayout(Const val: TGLTextLayout);
Begin
   FLayout := val;
   StructureChanged;
End;

Procedure TGLFlatText.SetModulateColor(Const val: TGLColor);
Begin
   FModulateColor.Assign(val);
End;

Procedure TGLFlatText.SetOptions(Const val: TGLFlatTextOptions);
Begin
   If val <> FOptions Then
   Begin
      FOptions := val;
      StructureChanged;
   End;
End;

Procedure TGLFlatText.DoRender(Var rci: TGLRenderContextInfo; aRenderSelf, aRenderChildren: Boolean);
Begin
   If Assigned(FBitmapFont) And (Text <> '') Then
   Begin
      rci.GLStates.PolygonMode := pmFill;
      If FModulateColor.Alpha <> 1 Then
      Begin
         rci.GLStates.Enable(stBlend);
         rci.GLStates.SetBlendFunc(bfSrcAlpha, bfOneMinusSrcAlpha);
      End;
      If ftoTwoSided In FOptions Then
         rci.GLStates.Disable(stCullFace);
      FBitmapFont.RenderString(rci, Text, FAlignment, FLayout,
         FModulateColor.Color);
   End;
   If Count > 0 Then
      Self.RenderChildren(0, Count - 1, rci);
End;

Procedure TGLFlatText.Assign(Source: TPersistent);
Begin
   If Assigned(Source) And (Source Is TGLFlatText) Then
   Begin
      BitmapFont := TGLFlatText(Source).BitmapFont;
      Text := TGLFlatText(Source).Text;
      Alignment := TGLFlatText(Source).Alignment;
      Layout := TGLFlatText(Source).Layout;
      ModulateColor := TGLFlatText(Source).ModulateColor;
      Options := TGLFlatText(Source).Options;
   End;
   Inherited Assign(Source);
End;


Procedure TGLCustomBitmapFont.SetCharRects(index: Integer; Const p: TVector);  //=== ct9999 ===
Begin
   If Length(FCharRects) < 256 Then
      Setlength(FCharRects, 256);
   FCharRects[index] := p;
End;




Initialization




   // class registrations
   RegisterClasses([TGLBitmapFont, TGLFlatText]);
End.
